"""The CLI tool options mix-ins."""

import os
import pytz

from plaso.analysis import manager as analysis_manager
from plaso.analyzers.hashers import manager as hashers_manager
from plaso.cli import logger
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.cli.helpers import profiling
from plaso.formatters import yaml_formatters_file
from plaso.lib import errors
from plaso.output import manager as output_manager


# TODO: pass argument_parser instead of argument_group and add groups
# in mix-ins.


class AnalysisPluginOptions:
  """Analysis plugin options mix-in."""

  # pylint: disable=no-member

  def _CreateAnalysisPlugins(self, options):
    """Creates the analysis plugins.

    Args:
      options (argparse.Namespace): command line arguments.

    Returns:
      dict[str, AnalysisPlugin]: analysis plugins and their names.
    """
    if not self._analysis_plugins:
      return {}

    analysis_plugins = (
        analysis_manager.AnalysisPluginManager.GetPluginObjects(
            self._analysis_plugins))

    for analysis_plugin in analysis_plugins.values():
      helpers_manager.ArgumentHelperManager.ParseOptions(
          options, analysis_plugin)

    return analysis_plugins

  def ListAnalysisPlugins(self):
    """Lists the analysis modules."""
    analysis_plugin_info = (
        analysis_manager.AnalysisPluginManager.GetAllPluginInformation())

    column_width = 10
    for name, _, _ in analysis_plugin_info:
      column_width = max(column_width, len(name))

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Analysis Plugins')
    # TODO: add support for a 3 column table.
    for name, description, type_string in analysis_plugin_info:
      table_view.AddRow([name, f'{description:s} [{type_string:s}]'])
    table_view.Write(self._output_writer)


class HashersOptions:
  """Hashers options mix-in."""

  # pylint: disable=no-member

  def __init__(self):
    """Initializes hasher options."""
    super().__init__()
    self._hasher_file_size_limit = None
    self._hasher_names_string = None

  def ListHashers(self):
    """Lists information about the available hashers."""
    hashers_information = hashers_manager.HashersManager.GetHashersInformation()

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Hashers')

    for name, description in sorted(hashers_information):
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)


class OutputModuleOptions:
  """Output module options mix-in.

  Attributes:
    list_time_zones (bool): True if the time zones should be listed.
  """

  # pylint: disable=no-member

  # Output format that have second-only date and time value and/or a limited
  # predefined set of output fields.
  _DEPRECATED_OUTPUT_FORMATS = frozenset(['l2tcsv', 'l2ttln', 'tln'])

  def __init__(self):
    """Initializes output module options."""
    super().__init__()
    self._output_additional_fields = []
    self._output_custom_fields = []
    self._output_custom_formatters_path = None
    self._output_dynamic_time = None
    self._output_filename = None
    self._output_format = None
    self._output_module = None
    self._output_time_zone = None

    self.list_time_zones = False

  def _CreateOutputModule(self, options):
    """Creates an output module.

    Args:
      options (argparse.Namespace): command line arguments.

    Returns:
      OutputModule: output module.

    Raises:
      BadConfigOption: if parameters are missing.
      RuntimeError: if the output module cannot be created.
    """
    if self._output_format in self._DEPRECATED_OUTPUT_FORMATS:
      self._PrintUserWarning((
          f'the output format: {self._output_format:s} has significant '
          f'limitations such as second-only date and time values and/or '
          f'a limited predefined set of output fields. It is strongly '
          f'recommend to use an alternative output format like: dynamic.'))

    try:
      output_module = output_manager.OutputManager.NewOutputModule(
          self._output_format)

    except (KeyError, ValueError) as exception:
      raise RuntimeError(
          f'Unable to create output module with error: {exception!s}')

    if output_module.WRITES_OUTPUT_FILE:
      if not self._output_filename:
        raise errors.BadConfigOption(
            f'Output format: {self._output_format:s} requires an output file')

      if os.path.exists(self._output_filename):
        raise errors.BadConfigOption(
            f'Output file already exists: {self._output_filename:s}')

      output_module.Open(path=self._output_filename)
    else:
      output_module.Open()

    helpers_manager.ArgumentHelperManager.ParseOptions(options, output_module)

    # Check if there are parameters that have not been defined and need to
    # in order for the output module to continue. Prompt user to supply
    # those that may be missing.
    missing_parameters = output_module.GetMissingArguments()
    if missing_parameters and self._unattended_mode:
      parameters_string = ', '.join(missing_parameters)
      raise errors.BadConfigOption((
          f'Unable to create output module missing parameters: '
          f'{parameters_string:s}'))

    while missing_parameters:
      self._PromptUserForMissingOutputModuleParameters(
          options, missing_parameters)

      helpers_manager.ArgumentHelperManager.ParseOptions(options, output_module)
      missing_parameters = output_module.GetMissingArguments()

    if output_module.SUPPORTS_ADDITIONAL_FIELDS:
      output_module.SetAdditionalFields(self._output_additional_fields)
    elif self._output_additional_fields:
      self._PrintUserWarning((
          f'output module: {self._output_format:s} does not support '
          f'additional fields'))

    if output_module.SUPPORTS_CUSTOM_FIELDS:
      output_module.SetCustomFields(self._output_custom_fields)
    elif self._output_custom_fields:
      self._PrintUserWarning((
          f'output module: {self._output_format:s} does not support '
          f'custom fields'))

    return output_module

  def _GetOutputModulesInformation(self):
    """Retrieves the output modules information.

    Returns:
      list[tuple[str, str]]: pairs of output module names and descriptions.
    """
    output_modules_information = []
    for name, output_class in output_manager.OutputManager.GetOutputClasses():
      output_modules_information.append((name, output_class.DESCRIPTION))

    return output_modules_information

  def _ParseOutputOptions(self, options):
    """Parses the output options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    additional_fields = self.ParseStringOption(options, 'additional_fields')
    if additional_fields:
      self._output_additional_fields = additional_fields.split(',')

    custom_fields = self.ParseStringOption(options, 'custom_fields')
    if custom_fields:
      for custom_field in custom_fields.split(','):
        try:
          name, value = custom_field.split(':')
        except ValueError:
          raise errors.BadConfigOption(
              f'Unsupported custom field: {custom_field:s}')

        self._output_custom_fields.append((name, value))

    custom_formatters_path = self.ParseStringOption(
        options, 'custom_formatter_definitions_path')

    if custom_formatters_path and not os.path.isfile(custom_formatters_path):
      raise errors.BadConfigOption((
          f'Unable to determine path to custom formatter definitions: '
          f'{custom_formatters_path:s}'))

    if custom_formatters_path:
      logger.info(
          f'Custom formatter definitions path: {custom_formatters_path:s}')

    if custom_formatters_path:
      message_formatters_file = yaml_formatters_file.YAMLFormattersFile()
      message_formatters_file.ReadFromFile(custom_formatters_path)

    self._output_custom_formatters_path = custom_formatters_path

    self._output_dynamic_time = getattr(options, 'dynamic_time', False)

    time_zone_string = self.ParseStringOption(options, 'output_time_zone')
    if isinstance(time_zone_string, str):
      if time_zone_string.lower() == 'list':
        self.list_time_zones = True

      elif time_zone_string:
        try:
          pytz.timezone(time_zone_string)
        except pytz.UnknownTimeZoneError:
          raise errors.BadConfigOption(
              f'Unknown time zone: {time_zone_string:s}')

        self._output_time_zone = time_zone_string

  def _PromptUserForMissingOutputModuleParameters(
      self, options, missing_parameters):
    """Prompts the user for missing output module parameters.

    Args:
      options (argparse.Namespace): command line arguments.
      missing_parameters (list[str]): names of missing output module parameters.
    """
    for parameter in missing_parameters:
      value = None
      while not value:
        value = self._PromptUserForInput(
            f'Please provide a value for {parameter:s}')

      setattr(options, parameter, value)

  def AddOutputOptions(self, argument_group):
    """Adds the output options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    argument_group.add_argument(
        '--additional_fields', '--additional-fields', dest='additional_fields',
        type=str, action='store', default='', help=(
            'Defines additional fields to be included in the output besides '
            'the default fields. Multiple additional field names can be '
            'defined as a list of comma separated values. Output formats that '
            'support additional fields are: dynamic, opensearch and xlsx.'))

    argument_group.add_argument(
        '--custom_fields', '--custom-fields', dest='custom_fields',
        type=str, action='store', default='', help=(
            'Defines custom fields to be included in the output besides '
            'the default fields. A custom field is defined as "name:value". '
            'Multiple custom field names can be defined as list of comma '
            'separated values. Note that regular fields will are favoured '
            'above custom fields with same name. Output formats that support '
            'this are: dynamic, opensearch and xlsx.'))

    argument_group.add_argument(
        '--custom_formatter_definitions', '--custom-formatter-definitions',
        dest='custom_formatter_definitions_path', type=str, metavar='PATH',
        action='store', help=(
            'Path to a file containing custom event formatter definitions, '
            'which is a .yaml file. Custom event formatter definitions can '
            'be used to customize event messages and override the built-in '
            'event formatter definitions.'))

    argument_group.add_argument(
        '--dynamic_time', '--dynamic-time', dest='dynamic_time',
        action='store_true', default=False, help=(
            'Indicate that the output should use dynamic time. Output formats '
            'that support dynamic time are: dynamic'))

    # Note the default here is None so we can determine if the time zone
    # option was set.
    argument_group.add_argument(
        '--output_time_zone', '--output-time-zone', dest='output_time_zone',
        action='store', metavar='TIME_ZONE', type=str, default=None, help=(
            'time zone of date and time values written to the output, if '
            'supported by the output format. Use "list" to see a list of '
            'available time zones. Output formats that support an output '
            'time zone are: dynamic and l2t_csv.'))

  def ListOutputModules(self):
    """Lists the output modules."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Output Modules')

    output_classes = sorted(
        output_manager.OutputManager.GetOutputClasses())
    for name, output_class in output_classes:
      table_view.AddRow([name, output_class.DESCRIPTION])
    table_view.Write(self._output_writer)

    disabled_classes = sorted(
        output_manager.OutputManager.GetDisabledOutputClasses())
    if not disabled_classes:
      return

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Disabled Output Modules')
    for name, output_class in disabled_classes:
      table_view.AddRow([name, output_class.DESCRIPTION])
    table_view.Write(self._output_writer)


class ProfilingOptions:
  """Profiling options mix-in."""

  # pylint: disable=no-member

  def __init__(self):
    """Initializes profiling options."""
    super().__init__()
    self._profilers = set()
    self._profiling_directory = None
    self._profiling_sample_rate = (
        profiling.ProfilingArgumentsHelper.DEFAULT_PROFILING_SAMPLE_RATE)

  def ListProfilers(self):
    """Lists information about the available profilers."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Profilers')

    profilers_information = sorted(
        profiling.ProfilingArgumentsHelper.PROFILERS_INFORMATION.items())
    for name, description in profilers_information:
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)


class StorageFileOptions:
  """Storage file options mix-in."""

  def AddStorageOptions(self, argument_parser):
    """Adds the storage options to the argument group.

    Args:
      argument_parser (argparse.ArgumentParser): argparse argument parser.
    """
    argument_parser.add_argument(
        'storage_file', metavar='PATH', nargs='?', type=str, default=None,
        help='Path to a storage file.')



"""SQLite parser plugin for iOS Notes database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class IOSNotesEventData(events.EventData):
  """iOS Notes event data.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): date and time the note was
        created.
    modification_time (dfdatetime.DateTimeValues): date and time the note was
        last modified.
    snippet (str): snippet of the note.
    title (str): title of the note.
  """

  DATA_TYPE = 'ios:notes:note'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.creation_time = None
    self.modification_time = None
    self.snippet = None
    self.title = None


class IOSNotesPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS Notes database files."""

  NAME = 'ios_notes'
  DATA_FORMAT = 'iOS Notes SQLite database file'

  REQUIRED_STRUCTURE = {
      'ZICCLOUDSYNCINGOBJECT': frozenset([
          'ZCREATIONDATE3', 'ZMODIFICATIONDATE1', 'ZSNIPPET', 'ZTITLE1'])}

  QUERIES = [
      (('SELECT ZCREATIONDATE3, ZMODIFICATIONDATE1, ZSNIPPET, ZTITLE1 '
        'FROM ZICCLOUDSYNCINGOBJECT'), '_ParseNoteRow')]

  SCHEMAS = [{
      'ACHANGE': (
          'CREATE TABLE ACHANGE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZCHANGETYPE INTEGER, ZENTITY INTEGER, ZENTITYPK '
          'INTEGER, ZTRANSACTIONID INTEGER, ZCOLUMNS BLOB )'),
      'ATRANSACTION': (
          'CREATE TABLE ATRANSACTION ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZAUTHORTS INTEGER, ZBUNDLEIDTS INTEGER, '
          'ZCONTEXTNAMETS INTEGER, ZPROCESSIDTS INTEGER, ZTIMESTAMP FLOAT, '
          'ZAUTHOR VARCHAR, ZBUNDLEID VARCHAR, ZCONTEXTNAME VARCHAR, '
          'ZPROCESSID VARCHAR, ZQUERYGEN BLOB )'),
      'ATRANSACTIONSTRING': (
          'CREATE TABLE ATRANSACTIONSTRING ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZNAME VARCHAR )'),
      'ZICCLOUDSTATE': (
          'CREATE TABLE ZICCLOUDSTATE ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZCURRENTLOCALVERSION INTEGER, ZINCLOUD '
          'INTEGER, ZLATESTVERSIONSYNCEDTOCLOUD INTEGER, ZCLOUDSYNCINGOBJECT '
          'INTEGER, Z2_CLOUDSYNCINGOBJECT INTEGER, ZLOCALVERSIONDATE '
          'TIMESTAMP )'),
      'ZICCLOUDSYNCINGOBJECT': (
          'CREATE TABLE ZICCLOUDSYNCINGOBJECT ( Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZCRYPTOITERATIONCOUNT INTEGER, '
          'ZISPASSWORDPROTECTED INTEGER, ZMARKEDFORDELETION INTEGER, '
          'ZMINIMUMSUPPORTEDNOTESVERSION INTEGER, ZNEEDSINITIALFETCHFROMCLOUD '
          'INTEGER, ZNEEDSTOBEFETCHEDFROMCLOUD INTEGER, '
          'ZNEEDSTOSAVEUSERSPECIFICRECORD INTEGER, ZCLOUDSTATE INTEGER, '
          'ZACCOUNT INTEGER, ZCHECKEDFORLOCATION INTEGER, ZFILESIZE INTEGER, '
          'ZHANDWRITINGSUMMARYVERSION INTEGER, ZHASMARKUPDATA INTEGER, '
          'ZIMAGECLASSIFICATIONSUMMARYVERSION INTEGER, ZIMAGEFILTERTYPE '
          'INTEGER, ZNEEDSINITIALRELATIONSHIPSETUP INTEGER, '
          'ZOCRSUMMARYVERSION INTEGER, ZORIENTATION INTEGER, ZSECTION '
          'INTEGER, ZURLEXPIRED INTEGER, ZLOCATION INTEGER, ZMEDIA INTEGER, '
          'ZNOTE INTEGER, ZNOTEUSINGTITLEFORNOTETITLE INTEGER, '
          'ZPARENTATTACHMENT INTEGER, ZAPPEARANCETYPE INTEGER, '
          'ZSCALEWHENDRAWING INTEGER, ZVERSION INTEGER, ZVERSIONOUTOFDATE '
          'INTEGER, ZATTACHMENT INTEGER, ZSTATE INTEGER, ZACCOUNT1 INTEGER, '
          'ZACCOUNT2 INTEGER, ZMENTIONNOTIFICATIONATTEMPTCOUNT INTEGER, '
          'ZMENTIONNOTIFICATIONSTATE INTEGER, ZNOTE1 INTEGER, '
          'ZPARENTATTACHMENT1 INTEGER, ZTYPE INTEGER, ZACCOUNT3 INTEGER, '
          'ZATTACHMENT1 INTEGER, ZATTACHMENTVIEWTYPE INTEGER, ZISPINNED '
          'INTEGER, ZISSYSTEMPAPER INTEGER, ZLEGACYNOTEWASPLAINTEXT INTEGER, '
          'ZNOTEHASCHANGES INTEGER, ZPAPERSTYLETYPE INTEGER, '
          'ZPREFERREDBACKGROUNDTYPE INTEGER, ZACCOUNT4 INTEGER, ZFOLDER '
          'INTEGER, ZNOTEDATA INTEGER, ZTITLESOURCEATTACHMENT INTEGER, '
          'ZISHIDDENNOTECONTAINER INTEGER, ZSORTORDER INTEGER, ZOWNER '
          'INTEGER, ZACCOUNTTYPE INTEGER, ZDIDCHOOSETOMIGRATE INTEGER, '
          'ZDIDFINISHMIGRATION INTEGER, ZDIDMIGRATEONMAC INTEGER, '
          'ZSERVERSIDEUPDATETASKFAILURECOUNT INTEGER, ZSTOREDATASEPARATELY '
          'INTEGER, ZACCOUNTDATA INTEGER, ZCUSTOMNOTESORTTYPEVALUE INTEGER, '
          'ZFOLDERTYPE INTEGER, ZIMPORTEDFROMLEGACY INTEGER, ZACCOUNT5 '
          'INTEGER, ZPARENT INTEGER, ZCREATIONDATE TIMESTAMP, '
          'ZCROPPINGQUADBOTTOMLEFTX FLOAT, ZCROPPINGQUADBOTTOMLEFTY FLOAT, '
          'ZCROPPINGQUADBOTTOMRIGHTX FLOAT, ZCROPPINGQUADBOTTOMRIGHTY FLOAT, '
          'ZCROPPINGQUADTOPLEFTX FLOAT, ZCROPPINGQUADTOPLEFTY FLOAT, '
          'ZCROPPINGQUADTOPRIGHTX FLOAT, ZCROPPINGQUADTOPRIGHTY FLOAT, '
          'ZDURATION FLOAT, ZMODIFICATIONDATE TIMESTAMP, ZORIGINX FLOAT, '
          'ZORIGINY FLOAT, ZPREVIEWUPDATEDATE TIMESTAMP, ZSIZEHEIGHT FLOAT, '
          'ZSIZEWIDTH FLOAT, ZHEIGHT FLOAT, ZMODIFIEDDATE TIMESTAMP, ZSCALE '
          'FLOAT, ZWIDTH FLOAT, ZSTATEMODIFICATIONDATE TIMESTAMP, '
          'ZCREATIONDATE1 TIMESTAMP, ZCREATIONDATE2 TIMESTAMP, '
          'ZMODIFICATIONDATEATIMPORT TIMESTAMP, ZCREATIONDATE3 TIMESTAMP, '
          'ZFOLDERMODIFICATIONDATE TIMESTAMP, '
          'ZLASTACTIVITYRECENTUPDATESVIEWEDDATE TIMESTAMP, '
          'ZLASTACTIVITYSUMMARYVIEWEDDATE TIMESTAMP, '
          'ZLASTATTRIBUTIONSVIEWEDDATE TIMESTAMP, ZLASTNOTIFIEDDATE '
          'TIMESTAMP, ZLASTOPENEDDATE TIMESTAMP, ZLASTVIEWEDMODIFICATIONDATE '
          'TIMESTAMP, ZLEGACYMODIFICATIONDATEATIMPORT TIMESTAMP, '
          'ZMODIFICATIONDATE1 TIMESTAMP, ZCUSTOMNOTESORTTYPEMODIFICATIONDATE '
          'TIMESTAMP, ZDATEFORLASTTITLEMODIFICATION TIMESTAMP, '
          'ZPARENTMODIFICATIONDATE TIMESTAMP, ZIDENTIFIER VARCHAR, '
          'ZPASSWORDHINT VARCHAR, ZZONEOWNERNAME VARCHAR, '
          'ZADDITIONALINDEXABLETEXT VARCHAR, ZFALLBACKSUBTITLEIOS VARCHAR, '
          'ZFALLBACKSUBTITLEMAC VARCHAR, ZFALLBACKTITLE VARCHAR, '
          'ZHANDWRITINGSUMMARY VARCHAR, ZIMAGECLASSIFICATIONSUMMARY VARCHAR, '
          'ZOCRSUMMARY VARCHAR, ZREMOTEFILEURLSTRING VARCHAR, ZSUMMARY '
          'VARCHAR, ZTITLE VARCHAR, ZTYPEUTI VARCHAR, ZURLSTRING VARCHAR, '
          'ZUSERTITLE VARCHAR, ZDEVICEIDENTIFIER VARCHAR, ZDISPLAYTEXT '
          'VARCHAR, ZSTANDARDIZEDCONTENT VARCHAR, ZALTTEXT VARCHAR, '
          'ZTOKENCONTENTIDENTIFIER VARCHAR, ZTYPEUTI1 VARCHAR, '
          'ZCONTENTHASHATIMPORT VARCHAR, ZFILENAME VARCHAR, '
          'ZLEGACYCONTENTHASHATIMPORT VARCHAR, ZLEGACYIMPORTDEVICEIDENTIFIER '
          'VARCHAR, ZLEGACYMANAGEDOBJECTIDURIREPRESENTATION VARCHAR, '
          'ZSELECTEDINKCOLORSTRING VARCHAR, ZSELECTEDINKIDENTIFIER VARCHAR, '
          'ZSNIPPET VARCHAR, ZTHUMBNAILATTACHMENTIDENTIFIER VARCHAR, ZTITLE1 '
          'VARCHAR, ZACCOUNTNAMEFORACCOUNTLISTSORTING VARCHAR, '
          'ZNESTEDTITLEFORSORTING VARCHAR, ZNAME VARCHAR, '
          'ZSERVERSIDEUPDATETASKLASTATTEMPTEDBUILD VARCHAR, '
          'ZSERVERSIDEUPDATETASKLASTATTEMPTEDVERSION VARCHAR, '
          'ZSERVERSIDEUPDATETASKLASTCOMPLETEDBUILD VARCHAR, '
          'ZSERVERSIDEUPDATETASKLASTCOMPLETEDVERSION VARCHAR, ZUSERRECORDNAME '
          'VARCHAR, ZSMARTFOLDERQUERYJSON VARCHAR, ZTITLE2 VARCHAR, '
          'ZPAPERASSETSURL VARCHAR, ZPAPERDATABASEURL VARCHAR, '
          'ZREPLICAIDTOBUNDLEIDENTIFIER BLOB, ZACTIVITYEVENTSDATA BLOB, '
          'ZASSETCRYPTOINITIALIZATIONVECTOR BLOB, ZASSETCRYPTOTAG BLOB, '
          'ZCRYPTOINITIALIZATIONVECTOR BLOB, ZCRYPTOSALT BLOB, ZCRYPTOTAG '
          'BLOB, ZCRYPTOWRAPPEDKEY BLOB, ZENCRYPTEDVALUESJSON BLOB, '
          'ZREPLICAIDTONOTESVERSIONDATA BLOB, ZSERVERRECORDDATA BLOB, '
          'ZSERVERSHAREDATA BLOB, ZUNAPPLIEDENCRYPTEDRECORD BLOB, '
          'ZUSERSPECIFICSERVERRECORDDATA BLOB, ZMERGEABLEDATA BLOB, '
          'ZFALLBACKIMAGECRYPTOINITIALIZATIONVECTOR BLOB, '
          'ZFALLBACKIMAGECRYPTOTAG BLOB, ZLINKPRESENTATIONARCHIVEDMETADATA '
          'BLOB, ZMARKUPMODELDATA BLOB, ZMERGEABLEDATA1 BLOB, ZMETADATADATA '
          'BLOB, ZSYNAPSEDATA BLOB, ZCRYPTOMETADATAINITIALIZATIONVECTOR BLOB, '
          'ZCRYPTOMETADATATAG BLOB, ZENCRYPTEDMETADATA BLOB, ZMETADATA BLOB, '
          'ZLASTNOTIFIEDTIMESTAMPDATA BLOB, ZLASTVIEWEDTIMESTAMPDATA BLOB, '
          'ZREPLICAIDTOUSERIDDICTDATA BLOB, ZCRYPTOVERIFIER BLOB, '
          'ZSERVERSIDEUPDATETASKCONTINUATIONTOKEN BLOB, ZMERGEABLEDATA2 BLOB '
          ')'),
      'ZICLOCATION': (
          'CREATE TABLE ZICLOCATION ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZPLACEUPDATED INTEGER, ZATTACHMENT '
          'INTEGER, ZLATITUDE FLOAT, ZLONGITUDE FLOAT, ZPLACEMARKDATA BLOB )'),
      'ZICNOTEDATA': (
          'CREATE TABLE ZICNOTEDATA ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZNOTE INTEGER, ZCRYPTOINITIALIZATIONVECTOR '
          'BLOB, ZCRYPTOTAG BLOB, ZDATA BLOB )'),
      'ZICSERVERCHANGETOKEN': (
          'CREATE TABLE ZICSERVERCHANGETOKEN ( Z_PK INTEGER PRIMARY KEY, '
          'Z_ENT INTEGER, Z_OPT INTEGER, ZDATABASESCOPE INTEGER, ZACCOUNT '
          'INTEGER, ZOWNERNAME VARCHAR, ZZONENAME VARCHAR, '
          'ZCKSERVERCHANGETOKENDATA BLOB )'),
      'Z_METADATA': (
          'CREATE TABLE Z_METADATA (Z_VERSION INTEGER PRIMARY KEY, Z_UUID '
          'VARCHAR(255), Z_PLIST BLOB)'),
      'Z_MODELCACHE': (
          'CREATE TABLE Z_MODELCACHE (Z_CONTENT BLOB)'),
      'Z_PRIMARYKEY': (
          'CREATE TABLE Z_PRIMARYKEY (Z_ENT INTEGER PRIMARY KEY, Z_NAME '
          'VARCHAR, Z_SUPER INTEGER, Z_MAX INTEGER)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely
        identifies the query that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  def _ParseNoteRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a note row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between
          parsers and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = IOSNotesEventData()
    event_data.creation_time = self._GetDateTimeRowValue(
        query_hash, row, 'ZCREATIONDATE3')
    event_data.modification_time = self._GetDateTimeRowValue(
        query_hash, row, 'ZMODIFICATIONDATE1')
    event_data.title = self._GetRowValue(query_hash, row, 'ZTITLE1')
    event_data.snippet = self._GetRowValue(query_hash, row, 'ZSNIPPET')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(IOSNotesPlugin)



"""Output module that saves data into a JSON line format.

JSON line format is a single JSON entry or event per line instead of grouping
all the output into a single JSON entity.
"""

import json

from plaso.output import manager
from plaso.output import shared_json


class JSONLineOutputModule(shared_json.SharedJSONOutputModule):
  """Output module for the JSON line format."""

  NAME = 'json_line'
  DESCRIPTION = 'Saves the events into a JSON line format.'

  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """
    json_string = json.dumps(field_values, sort_keys=True)
    self.WriteLine(json_string)


manager.OutputManager.RegisterOutput(JSONLineOutputModule)



"""Null device output module."""

from plaso.output import interface
from plaso.output import manager


class NullOutputModule(interface.OutputModule):
  """Null device output module."""

  NAME = 'null'
  DESCRIPTION = 'Output module that does not output anything.'

  # pylint: disable=unused-argument

  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      dict[str, str]: output field values per name.
    """
    return {}

  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """
    return


manager.OutputManager.RegisterOutput(NullOutputModule)



"""Parser for iOS Discord message (JSON) files."""

import codecs
import json
import os

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import manager


class IOSDiscordMessageEventData(events.EventData):
  """iOS discord message event data.

  Attributes:
    attachment_name (str): The attachment filename.
    attachment_proxy_urls (str): The attachment proxy URL.
    attachment_size (int): The attachment size.
    attachment_type (str): The attachment type.
    channel_identifier (str): identifier of the user channel.
    content (str): Message content.
    edited_timestamp (str): Message edit time.
    sent_time (dfdatetime.DateTimeValues): Message timestamp.
    user_identifier (str): ID of the message author.
    username (str): The username of the message sender.
  """

  DATA_TYPE = 'ios:discord:message'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.attachment_name = None
    self.attachment_proxy_url = None
    self.attachment_size = None
    self.attachment_type = None
    self.channel_identifier = None
    self.content = None
    self.sent_time = None
    self.user_identifier = None
    self.username = None


class IOSDiscordParser(interface.FileObjectParser):
  """Parses iOS discord message files."""

  NAME = 'discord_ios'
  DATA_FORMAT = 'iOS discord message'

  REQUIRED_MESSAGE_KEYS = frozenset([
      'attachments', 'author', 'channel_id', 'content', 'timestamp'])

  _ENCODING = 'utf-8'
  _MAXIMUM_FILE_SIZE = 16 * 1024 * 1024

  def _GetDateTimeValue(self, parser_mediator, timestamp):
    """Retrieves a date and time value.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      timestamp (Optional[str]): timestamp value.

    Returns:
      dfdatetime.TimeElementsInMicroseconds: date and time value or None if
          not available.
    """
    if timestamp is None:
      return None

    try:
      date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()
      date_time.CopyFromStringISO8601(timestamp)
    except ValueError as exception:
      parser_mediator.ProduceExtractionWarning(
          f'Unable to parse timestamp value: {timestamp:s} with error: '
          f'{exception!s}')
      date_time = None

    return date_time

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a iOS discord message file."""
    # First check for initial 2 characters being open brace and open list.
    if file_object.read(2) != b'[{':
      display_name = parser_mediator.GetDisplayName()
      raise errors.WrongParser(
          f'[{self.NAME!s}] {display_name!s} is not a valid Discord messages '
          f'file, missing opening brace and list')

    file_object.seek(0, os.SEEK_SET)

    # Note that _MAXIMUM_FILE_SIZE prevents this read to become too large.
    file_content = file_object.read()

    try:
      file_content = codecs.decode(file_content, self._ENCODING)
    except (UnicodeDecodeError, json.JSONDecodeError):
      display_name = parser_mediator.GetDisplayName()
      raise errors.WrongParser(
          f'[{self.NAME!s}] {display_name!s} is not a valid Discord messages '
          f'file, unable to decode as UTF-8')

    # Second verify it is valid JSON.
    try:
      json_dict = json.loads(file_content)

    except OSError as exception:
      display_name = parser_mediator.GetDisplayName()
      raise errors.WrongParser(
          f'[{self.NAME!s}] Unable to open file {display_name!s} with error: '
          f'{exception!s}')

    except ValueError as exception:
      display_name = parser_mediator.GetDisplayName()
      raise errors.WrongParser(
          f'[{self.NAME!s}] {display_name!s} is not a valid Discord messages '
          f'file, unable to decode as JSON with error: {exception!s}')

    # Third verify the file has the correct keys for a Discord messages file.
    messages = json_dict or [{}]

    if not set(self.REQUIRED_MESSAGE_KEYS).issubset(set(messages[0].keys())):
      raise errors.WrongParser('File does not contain Discord messages data')

    for message in messages:
      attachments = message.get('attachments') or [{}]
      timestamp = message.get('timestamp')

      event_data = IOSDiscordMessageEventData()
      event_data.attachment_name = attachments[0].get('filename') or None
      event_data.attachment_proxy_url = attachments[0].get('proxy_url') or None
      event_data.attachment_size = attachments[0].get('size') or None
      event_data.attachment_type = attachments[0].get('content_type') or None
      event_data.channel_identifier = message.get('channel_id') or None
      event_data.content = message.get('content') or None
      event_data.sent_time = self._GetDateTimeValue(parser_mediator, timestamp)
      event_data.user_identifier = message.get('author', {}).get('id') or None
      event_data.username = message.get('author', {}).get('username') or None

      parser_mediator.ProduceEventData(event_data)


manager.ParsersManager.RegisterParser(IOSDiscordParser)



"""Plist parser plugin for MacOS user plist files.

Fields within the plist key:
  name: username.
  uid: user identifier (UID).
  passwordpolicyoptions: XML Plist structures with the timestamp.
  passwordLastSetTime: last time the password was changed.
  lastLoginTimestamp: last time the user was authenticated depending on
      the situation, these timestamps are reset (0 value). It is translated
      by the library as a 2001-01-01 00:00:00 (Cocoa zero time representation).
  failedLoginTimestamp: last time the login attempt failed.
  failedLoginCount: number of failed loging attempts.
"""

# TODO: Only plists from MacOS 10.8 and 10.9 were tested. Look at other
#       versions as well.

import codecs
import plistlib

from xml.parsers import expat

from defusedxml import ElementTree
from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import logger
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class MacOSUserEventData(events.EventData):
  """MacOS user event data.

  Attributes:
    fullname (str): full name.
    home_directory (str): path of the home directory.
    last_login_attempt_time (dfdatetime.DateTimeValues): date and time of
       the last (failed) login attempt.
    last_login_time (dfdatetime.DateTimeValues): date and time of the last
       login.
    last_password_set_time (dfdatetime.DateTimeValues): date and time of the
        last password set.
    number_of_failed_login_attempts (str): number of failed login attempts.
    password_hash (str): password hash.
    user_identifier (str): user identifier.
    username (str): username.
  """

  DATA_TYPE = 'macos:user:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.fullname = None
    self.home_directory = None
    self.last_login_attempt_time = None
    self.last_login_time = None
    self.last_password_set_time = None
    self.number_of_failed_login_attempts = None
    self.password_hash = None
    self.user_identifier = None
    self.username = None


class MacOSUserPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for MacOS user plist files."""

  NAME = 'macuser'
  DATA_FORMAT = 'MacOS user plist file'

  # The PLIST_PATH is dynamic, "user".plist is the name of the
  # MacOS user.
  PLIST_KEYS = frozenset([
      'name', 'uid', 'home', 'passwordpolicyoptions', 'ShadowHashData'])

  _ROOT = '/'

  def _GetDateTimeValueFromTimeString(
      self, parser_mediator, policy_values, value_name):
    """Retrieves a date and time value from a time string value.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      policy_values (dict[str, object]): policy values.
      value_name (str): name of the value.

    Returns:
      dfdatetime.TimeElements: date and time or None if not available.
    """
    time_string = policy_values.get(value_name)
    if not time_string or time_string == '2001-01-01T00:00:00Z':
      return None

    date_time = dfdatetime_time_elements.TimeElements()

    try:
      date_time.CopyFromStringISO8601(time_string)
    except (TypeError, ValueError):
      parser_mediator.ProduceExtractionWarning(
          f'unable to parse value: {value_name:s} time string: {time_string!s}')
      return None

    return date_time

  # pylint: disable=arguments-differ
  def _ParsePlist(
      self, parser_mediator, match=None, top_level=None, **unused_kwargs):
    """Extracts relevant user timestamp entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    if 'name' not in match or 'uid' not in match:
      return

    password_hash = None
    user_identifier = match['uid'][0]
    username = match['name'][0]

    shadow_hash_data = match.get('ShadowHashData')
    if isinstance(shadow_hash_data, (list, tuple)):
      # Extract the hash password information, which is stored in
      # the attribute ShadowHashData which is a binary plist data.
      try:
        property_list = plistlib.loads(shadow_hash_data[0])
      except plistlib.InvalidFileException as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to parse ShadowHashData with error: {exception!s}')
        property_list = {}

      salted_hash = property_list.get('SALTED-SHA512-PBKDF2')
      if salted_hash:
        salt_hex_bytes = codecs.encode(salted_hash['salt'], 'hex')
        salt_string = codecs.decode(salt_hex_bytes, 'ascii')
        entropy_hex_bytes = codecs.encode(salted_hash['entropy'], 'hex')
        entropy_string = codecs.decode(entropy_hex_bytes, 'ascii')

        number_of_iterations = salted_hash["iterations"]
        password_hash = (
            f'$ml${number_of_iterations:d}${salt_string:s}${entropy_string:s}')

    for policy in match.get('passwordpolicyoptions', []):
      try:
        xml_policy = ElementTree.fromstring(policy)
      except (LookupError, ElementTree.ParseError,
              expat.ExpatError) as exception:
        logger.error(
            f'Unable to parse XML structure for an user policy, username: '
            f'{username:s} and UID: {user_identifier!s}, with error: '
            f'{exception!s}')
        continue

      for dict_elements in xml_policy.iterfind('dict'):
        key_values = [value.text for value in dict_elements]
        # Taking a list and converting it to a dict, using every other item
        # as the key and the other one as the value.
        policy_dict = dict(zip(key_values[0::2], key_values[1::2]))

      event_data = MacOSUserEventData()
      event_data.fullname = top_level.get('realname', [None])[0]
      event_data.home_directory = top_level.get('home', [None])[0]
      event_data.last_login_attempt_time = self._GetDateTimeValueFromTimeString(
          parser_mediator, policy_dict, 'failedLoginTimestamp')
      event_data.last_login_time = self._GetDateTimeValueFromTimeString(
          parser_mediator, policy_dict, 'lastLoginTimestamp')
      event_data.last_password_set_time = self._GetDateTimeValueFromTimeString(
          parser_mediator, policy_dict, 'passwordLastSetTime')
      event_data.number_of_failed_login_attempts = policy_dict.get(
          'failedLoginCount', None)
      event_data.password_hash = password_hash
      event_data.user_identifier = user_identifier
      event_data.username = username

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(MacOSUserPlistPlugin)



"""Parser for binary and text Property List (plist) files."""

import binascii
import plistlib

from xml.parsers import expat

from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import logger
from plaso.parsers import manager


class PlistParser(interface.FileObjectParser):
  """Parser for binary and text Property List (plist) files."""

  NAME = 'plist'
  DATA_FORMAT = 'Property list (plist) file'

  # 50 MB is 10x larger than any plist file seen to date.
  _MAXIMUM_FILE_SIZE = 50000000

  _UTF16BE_BYTE_ORDER_MARK = b'\xfe\xff'
  _UTF16LE_BYTE_ORDER_MARK = b'\xff\xfe'
  _UTF32BE_BYTE_ORDER_MARK = b'\x00\x00\xfe\xff'
  _UTF32LE_BYTE_ORDER_MARK = b'\xff\xfe\x00\x00'
  _UTF8_BYTE_ORDER_MARK = b'\xef\xbb\xbf'

  _plugin_classes = {}

  def _CheckByteOrderMark(self, plist_data):
    """Determines if the plist data starts with a byte-order-mark.

    Args:
      plist_data (bytes): plist data.

    Returns:
      tuple: containing:

        int: size of the byte-order-mark or 0 if no byte-order-mark was
            detected.
        str: encoding or ascii if no byte-order-mark was detected.
    """
    if plist_data.startswith(self._UTF32BE_BYTE_ORDER_MARK):
      return 4, 'utf-32-be'

    if plist_data.startswith(self._UTF32LE_BYTE_ORDER_MARK):
      return 4, 'utf-32-le'

    if plist_data.startswith(self._UTF16BE_BYTE_ORDER_MARK):
      return 2, 'utf-16-be'

    if plist_data.startswith(self._UTF16LE_BYTE_ORDER_MARK):
      return 2, 'utf-16-le'

    if plist_data.startswith(self._UTF8_BYTE_ORDER_MARK):
      return 3, 'utf-8'

    return 0, 'ascii'

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: a format specification or None if not available.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'bplist', offset=0)
    return format_specification

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a plist file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    filename = parser_mediator.GetFilename()

    # Note that _MAXIMUM_FILE_SIZE prevents this read to become too large.
    plist_data = file_object.read()

    has_leading_whitespace = False

    is_binary_plist = plist_data.startswith(b'bplist0')
    if not is_binary_plist:
      byte_order_mark_size, encoding = self._CheckByteOrderMark(plist_data)

      xml_signature = '<?xml '.encode(encoding)
      is_xml = plist_data[byte_order_mark_size:].startswith(xml_signature)
      if not is_xml:
        # Preserve the byte-order-mark for plistlib.
        plist_data = b''.join([
            plist_data[:byte_order_mark_size],
            plist_data[byte_order_mark_size:].lstrip()])
        is_xml = plist_data[byte_order_mark_size:].startswith(xml_signature)
        has_leading_whitespace = is_xml

      if is_xml:
        plist_footer = '</plist>'.encode(encoding)
        plist_data = plist_data.rstrip()

        if not plist_data.endswith(plist_footer):
          raise errors.WrongParser(
              'Unable to parse XML plist with error: missing plist XML root '
              'element.')

    try:
      top_level_object = plistlib.loads(plist_data)

    except (AttributeError, binascii.Error, expat.ExpatError,
            plistlib.InvalidFileException) as exception:
      raise errors.WrongParser(
          f'Unable to parse plist with error: {exception!s}')

    except (LookupError, ValueError) as exception:
      # LookupError will be raised in cases where the plist is an XML file
      # that contains an unsupported encoding.
      # ValueError will be raised in cases where the plist is an XML file
      # that contains an unsupported date and time value.
      parser_mediator.ProduceExtractionWarning(
          f'unable to parse XML plist with error: {exception!s}')
      return

    if not top_level_object:
      # Do not produce an extraction warning for a binary plist without a top
      # level object or an XML plist with an empty top level object.
      if top_level_object is None and not is_binary_plist:
        parser_mediator.ProduceExtractionWarning((
            'unable to parse XML plist file with error: missing top level '
            'object'))
      return

    if has_leading_whitespace:
      parser_mediator.ProduceExtractionWarning(
          'XML plist file with leading whitespace')

    display_name = parser_mediator.GetDisplayName()
    filename_lower_case = filename.lower()

    found_matching_plugin = False
    for plugin_name, plugin in self._plugins_per_name.items():
      if parser_mediator.abort:
        break

      profiling_name = '/'.join([self.NAME, plugin.NAME])

      parser_mediator.SampleFormatCheckStartTiming(profiling_name)

      try:
        if not plugin.PLIST_PATH_FILTERS:
          path_filter_match = True
        else:
          path_filter_match = False
          for path_filter in plugin.PLIST_PATH_FILTERS:
            if path_filter.Match(filename_lower_case):
              path_filter_match = True

        try:
          required_format = plugin.CheckRequiredFormat(top_level_object)
        except Exception as exception:  # pylint: disable=broad-except
          parser_mediator.ProduceExtractionWarning((
              f'plugin: {plugin_name:s} unable to parse plist file with error: '
              f'{exception!s}'))

      finally:
        parser_mediator.SampleFormatCheckStopTiming(profiling_name)

      if not path_filter_match or not required_format:
        logger.debug(f'Skipped parsing file: {display_name:s} '
                     f'with plugin: {plugin_name:s}')
        continue

      logger.debug(
          f'Parsing file: {display_name:s} with plugin: {plugin_name:s}')

      parser_mediator.SampleStartTiming(profiling_name)

      try:
        plugin.UpdateChainAndProcess(
            parser_mediator, top_level=top_level_object)
        found_matching_plugin = True

      except Exception as exception:  # pylint: disable=broad-except
        parser_mediator.ProduceExtractionWarning((
            f'plugin: {plugin_name:s} unable to parse plist file with error: '
            f'{exception!s}'))

      finally:
        parser_mediator.SampleStopTiming(profiling_name)

    if not found_matching_plugin and self._default_plugin:
      profiling_name = '/'.join([self.NAME, self._default_plugin.NAME])

      parser_mediator.SampleStartTiming(profiling_name)

      try:
        self._default_plugin.UpdateChainAndProcess(
            parser_mediator, top_level=top_level_object)
      finally:
        parser_mediator.SampleStopTiming(profiling_name)


manager.ParsersManager.RegisterParser(PlistParser)



"""SQLite parser plugin for MacOS Notification Center database files."""

import plistlib

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class MacOSNotificationCenterEventData(events.EventData):
  """MacOS NotificationCenter event data.

  Attributes:
    body (str): body of the notification message.
    bundle_name (str): name of the application's bundle that generated
        the notification.
    creation_time (dfdatetime.DateTimeValues): date and time the entry was
        created.
    presented (int): either 1 or 0 if the notification has been shown to the
        user.
    subtitle (str): optional. Subtitle of the notification message.
    title (str): title of the message. Usually the name of the application
        that generated the notification. Occasionally the name of the sender
        of the notification for example, in case of chat messages.
  """

  DATA_TYPE = 'macos:notification_center:entry'

  def __init__(self):
    """Initialize event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.bundle_name = None
    self.creation_time = None
    self.presented = None
    self.subtitle = None
    self.title = None


class MacOSNotificationCenterPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for MacOS Notification Center database files.

  The MacOS Notification Center database file is typically stored in:
  /private/var/folders/<W><d>/../0/com.apple.notificationcenter/db2/db

  At the moment it takes into consideration only the main table, 'record'.
  Currently supported tables and related content:
    Record: contains historical records
    Requests: contain pending requests
    Delivered: delivered requests
    Displayed: displayed requests, by app_id
    Snoozed: snoozed by user requests
  """

  NAME = 'mac_notificationcenter'
  DATA_FORMAT = 'MacOS Notification Center SQLite database file'

  REQUIRED_STRUCTURE = {
      'app': frozenset([
          'identifier', 'app_id']),
      'record': frozenset([
          'data', 'delivered_date', 'presented', 'app_id'])}

  QUERIES = [
      ('SELECT a.identifier AS bundle_name, '
       'r.data AS dataBlob, r.delivered_date AS timestamp,'
       'r.presented AS presented '
       'FROM app a, record r '
       'WHERE a.app_id = r.app_id', 'ParseNotificationcenterRow')]

  SCHEMAS = [{
      'app': (
          'CREATE TABLE app (app_id INTEGER PRIMARY KEY, identifier VARCHAR)'),
      'dbinfo': (
          'CREATE TABLE dbinfo (key VARCHAR, value VARCHAR)'),
      'delivered': (
          'CREATE TABLE delivered (app_id INTEGER PRIMARY KEY, list BLOB)'),
      'displayed': (
          'CREATE TABLE displayed (app_id INTEGER PRIMARY KEY, list BLOB)'),
      'record': (
          'CREATE TABLE record (rec_id INTEGER PRIMARY KEY, app_id INTEGER, '
          'uuid BLOB, data BLOB, request_date REAL, request_last_date REAL, '
          'delivered_date REAL, presented Bool, style INTEGER, '
          'snooze_fire_date REAL)'),
      'requests': (
          'CREATE TABLE requests (app_id INTEGER PRIMARY KEY, list BLOB)'),
      'snoozed': (
          'CREATE TABLE snoozed (app_id INTEGER PRIMARY KEY, list BLOB)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  def ParseNotificationcenterRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses a message row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    data_blob = self._GetRowValue(query_hash, row, 'dataBlob')

    try:
      property_list = plistlib.loads(data_blob)
      # req_property is the 'req' dictionary from the plist containing extra
      # information about the notification entry.
      req_property = property_list['req']

    except (KeyError, plistlib.InvalidFileException) as exception:
      parser_mediator.ProduceExtractionWarning(
          f'unable to read plist from database with error: {exception!s}')
      req_property = {}

    event_data = MacOSNotificationCenterEventData()
    event_data.body = req_property.get('body')
    event_data.bundle_name = self._GetRowValue(query_hash, row, 'bundle_name')
    event_data.creation_time = self._GetDateTimeRowValue(
        query_hash, row, 'timestamp')
    event_data.presented = self._GetRowValue(query_hash, row, 'presented')
    event_data.subtitle = req_property.get('subt')
    event_data.title = req_property.get('titl')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(MacOSNotificationCenterPlugin)



"""JSON-L parser plugin for Google Cloud (GCP) log files."""

import re

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import jsonl_parser
from plaso.parsers.jsonl_plugins import interface


class GCPLogEventData(events.EventData):
  """Google Cloud (GCP) log event data.

  Attributes:
    caller_ip (str): IP address of the client that requested the operation.
    container (str): TODO
    dcsa_emails (list[str]): default compute service account attached to a
        Google Compute Engine (GCE) instance.
    dcsa_scopes (list[str]): OAuth scopes granted to the default compute service
        account.
    delegation_chain (str): service account delegation chain.
    event_subtype (str): JSON event sub type or protocol buffer method.
    event_type (str): TODO
    filename (str): TODO
    firewall_rules (list[str]): firewall rules.
    firewall_source_ranges (list[str]): firewall source ranges.
    gcloud_command_identity (str): unique gcloud command identity.
    gcloud_command_partial (str): partial gcloud command.
    log_name (str): name of the log entry.
    message (str): TODO
    method_name (str): operation performed.
    permissions (list[str]): IAM permission used for the operation.
    policy_deltas (list[str]): TODO
    principal_email (str): email address of the requester.
    principal_subject (str): subject name of the requester.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
    request_account_identifier (str): GCP account identifier of the request.
    request_address (str): IP address assigned to a Google Cloud Engine (GCE)
        instance.
    request_description (str): description of the request.
    request_direction (str): direction of the request.
    request_email (str): email address of the request.
    request_member (str): member of the request.
    request_metadata (list[str]): request metadata values.
    request_name (str): name of the request.
    request_target_tags (str): TODO
    resource_labels (list[str]): resource labels.
    resource_name (str): name of the resource.
    service_account_delegation (list[str]): service accounts delegation in the
        authentication.
    service_account_display_name (str): display name of the service account.
    service_account_key_name (str): service account key name used in
        authentication.
    service_name (str): name of the service.
    severity (str): log entry severity.
    source_images (list[str]): source images of disks attached to a compute
        engine instance.
    status_code (str): operation success or failure code.
    status_message (str): operation success or failure message.
    status_reasons (list[str]): reasons for operation failure.
    text_payload (str): text payload for logs not using a JSON or proto payload.
    user_agent (str): user agent used in the request.
  """

  DATA_TYPE = 'gcp:log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.caller_ip = None
    self.container = None
    self.dcsa_emails = None
    self.dcsa_scopes = None
    self.delegation_chain = None
    self.event_subtype = None
    self.event_type = None
    self.filename = None
    self.firewall_rules = None
    self.firewall_source_ranges = None
    self.gcloud_command_identity = None
    self.gcloud_command_partial = None
    self.log_name = None
    self.method_name = None
    self.message = None
    self.permissions = None
    self.policy_deltas = None
    self.principal_email = None
    self.principal_subject = None
    self.recorded_time = None
    self.request_account_identifier = None
    self.request_address = None
    self.request_description = None
    self.request_direction = None
    self.request_email = None
    self.request_member = None
    self.request_metadata = None
    self.request_name = None
    self.request_target_tags = None
    self.resource_labels = None
    self.resource_name = None
    self.service_account_delegation = None
    self.service_account_display_name = None
    self.service_account_key_name = None
    self.service_name = None
    self.severity = None
    self.source_images = None
    self.status_code = None
    self.status_message = None
    self.status_reasons = None
    self.text_payload = None
    self.user_agent = None


class GCPLogJSONLPlugin(interface.JSONLPlugin):
  """JSON-L parser plugin for Google Cloud (GCP) log files."""

  NAME = 'gcp_log'
  DATA_FORMAT = 'Google Cloud (GCP) log'

  _USER_AGENT_COMMAND_RE = re.compile(r'command/([^\s]+)')

  _USER_AGENT_INVOCATION_ID_RE = re.compile(r'invocation-id/([^\s]+)')

  def _ParseJSONPayload(self, json_dict, event_data):
    """Extracts information from a jsonPayload value.

    Args:
      json_dict (dict): JSON dictionary of the log record.
      event_data (GCPLogEventData): event data.
    """
    json_payload = self._GetJSONValue(json_dict, 'jsonPayload')
    if not json_payload:
      return

    event_data.container = self._GetJSONValue(json_payload, 'container')
    event_data.event_subtype = self._GetJSONValue(json_payload, 'event_subtype')
    event_data.event_type = self._GetJSONValue(json_payload, 'event_type')
    event_data.filename = self._GetJSONValue(json_payload, 'filename')
    event_data.message = self._GetJSONValue(json_payload, 'message')

    actor_json = self._GetJSONValue(json_payload, 'actor')
    if actor_json:
      event_data.user = self._GetJSONValue(actor_json, 'user')

  def _ParseAuthenticationInfo(self, proto_payload, event_data):
    """Extracts information from `protoPayload.authenticationInfo`.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    authentication_info = self._GetJSONValue(
        proto_payload, 'authenticationInfo')
    if not authentication_info:
      return

    principal_email = self._GetJSONValue(authentication_info, 'principalEmail')
    if principal_email:
      event_data.principal_email = principal_email

    principal_subject = self._GetJSONValue(
        authentication_info, 'principalSubject')
    if principal_subject:
      event_data.principal_subject = principal_subject

    service_account_key_name = self._GetJSONValue(
        authentication_info, 'serviceAccountKeyName')
    if service_account_key_name:
      event_data.service_account_key_name = service_account_key_name

    delegations = []

    delegation_info_list = self._GetJSONValue(
        authentication_info, 'serviceAccountDelegationInfo', [])
    for delegation_info in delegation_info_list:
      first_party_principal = self._GetJSONValue(
          delegation_info, 'firstPartyPrincipal', {})

      first_party_principal_email = self._GetJSONValue(first_party_principal,
          'principalEmail')
      if first_party_principal_email:
        delegations.append(first_party_principal_email)
      else:
        first_party_principal_subject = self._GetJSONValue(
            first_party_principal, 'principalSubject')
        if first_party_principal_subject:
          delegations.append(first_party_principal_subject)

    if delegations:
      event_data.service_account_delegation = delegations
      event_data.delegation_chain = '->'.join(delegations)

  def _ParseAuthorizationInfo(self, proto_payload, event_data):
    """Extracts information from `protoPayload.authorizationInfo`.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    permissions = []

    authorization_info_list = self._GetJSONValue(
        proto_payload, 'authorizationInfo', [])
    for authorization_info in authorization_info_list:
      permission = self._GetJSONValue(authorization_info, 'permission')
      if permission:
        permissions.append(permission)

    if permissions:
      event_data.permissions = permissions

  def _ParseRequestMetadata(self, proto_payload, event_data):
    """Extracts information from `protoPayload.requestMetadata`.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    request_metadata = self._GetJSONValue(proto_payload, 'requestMetadata')
    if not request_metadata:
      return

    event_data.caller_ip = self._GetJSONValue(request_metadata, 'callerIp')
    event_data.user_agent = self._GetJSONValue(
        request_metadata, 'callerSuppliedUserAgent')

    if event_data.user_agent:
      if 'command/' in event_data.user_agent:
        matches = self._USER_AGENT_COMMAND_RE.search(event_data.user_agent)
        if matches:
          command_string = matches.group(1).replace('.', ' ')
          event_data.gcloud_command_partial = command_string

      if 'invocation-id' in event_data.user_agent:
        matches = self._USER_AGENT_INVOCATION_ID_RE.search(
            event_data.user_agent)
        if matches:
          event_data.gcloud_command_identity = matches.group(1)

  def _ParseProtoPayloadStatus(self, proto_payload, event_data):
    """Extracts information from `protoPayload.status`.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    status = self._GetJSONValue(proto_payload, 'status')
    if status:
      # Non empty `protoPayload.status` field could have empty
      # `protoPayload.status.code` field.
      #
      # Empty `code` and `message` fields indicate the operation was successful.

      event_data.status_code = str(self._GetJSONValue(status, 'code', ''))
      event_data.status_message = self._GetJSONValue(status, 'message')

      # `protoPayload.status.details[].reason` contains reason for an operation
      # failure.
      status_reasons = []

      for status_detail in self._GetJSONValue(status, 'details', []):
        status_reason = self._GetJSONValue(status_detail, 'reason')
        if status_reason:
          status_reasons.append(status_reason)

      if status_reasons:
        event_data.status_reasons = status_reasons

  def _ParseComputeInsertRequest(self, request, event_data):
    """Extracts compute.instances.insert information.

    Args:
      request (dict): JSON dictionary of the `protoPayload.request` field.
      event_data (GCPLogEventData): event data.
    """
    # source_images hold Google Cloud source disk path used in creating a GCE
    # instance.
    source_images = []

    for disk in self._GetJSONValue(request, 'disks', []):
      initialize_params = self._GetJSONValue(disk, 'initializeParams', {})

      source_image = self._GetJSONValue(initialize_params, 'sourceImage')
      if source_image:
        source_images.append(source_image)

    if source_images:
      event_data.source_images = source_images

    # Default compute service account aka dcsa
    dcsa_emails = []
    dcsa_scopes = []

    service_account_list = self._GetJSONValue(request, 'serviceAccounts', [])
    for service_account in service_account_list:
      email = self._GetJSONValue(service_account, 'email')
      if email:
        dcsa_emails.append(email)

      scopes = self._GetJSONValue(service_account, 'scopes')
      if scopes:
        dcsa_scopes.extend(scopes)

    if dcsa_emails:
      event_data.dcsa_emails = dcsa_emails

    if dcsa_scopes:
      event_data.dcsa_scopes = dcsa_scopes

  def _ParseComputeProtoPayload(self, proto_payload, event_data):
    """Extracts compute.googleapis.com information.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    request = self._GetJSONValue(proto_payload, 'request')
    if not request:
      return

    request_type = self._GetJSONValue(request, '@type')
    if not request_type:
      return

    if request_type == 'type.googleapis.com/compute.instances.insert':
      self._ParseComputeInsertRequest(request, event_data)

  def _ParseProtoPayload(self, json_dict, event_data):
    """Extracts information from a protoPayload value.

    Args:
      json_dict (dict): JSON dictionary of the log record.
      event_data (GCPLogEventData): event data.
    """
    proto_payload = self._GetJSONValue(json_dict, 'protoPayload')
    if not proto_payload:
      return

    event_data.service_name = self._GetJSONValue(proto_payload, 'serviceName')
    event_data.resource_name = self._GetJSONValue(proto_payload, 'resourceName')

    method_name = self._GetJSONValue(proto_payload, 'methodName')
    if method_name and not event_data.event_subtype:
      event_data.event_subtype = method_name
      event_data.method_name = method_name

    self._ParseAuthenticationInfo(proto_payload, event_data)
    self._ParseAuthorizationInfo(proto_payload, event_data)
    self._ParseRequestMetadata(proto_payload, event_data)
    self._ParseProtoPayloadStatus(proto_payload, event_data)
    self._ParseProtoPayloadRequest(proto_payload, event_data)
    self._ParseProtoPayloadServiceData(proto_payload, event_data)

    if event_data.service_name == 'compute.googleapis.com':
      self._ParseComputeProtoPayload(proto_payload, event_data)

  def _ParseProtoPayloadRequest(self, proto_payload, event_data):
    """Extracts information from the request field of a protoPayload field.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    request = self._GetJSONValue(proto_payload, 'request')
    if not request:
      return

    event_data.request_account_identifier = self._GetJSONValue(
        request, 'account_id')
    event_data.request_address = self._GetJSONValue(request, 'address')
    event_data.request_description = self._GetJSONValue(request, 'description')
    event_data.request_direction = self._GetJSONValue(request, 'direction')
    event_data.request_email = self._GetJSONValue(request, 'email')
    event_data.request_member = self._GetJSONValue(request, 'member')
    event_data.request_name = self._GetJSONValue(request, 'name')
    event_data.request_target_tags = self._GetJSONValue(request, 'targetTags')

    # Firewall specific attributes.
    event_data.firewall_source_ranges = self._GetJSONValue(
        request, 'sourceRanges')

    firewall_rules = []

    alloweds = self._GetJSONValue(request, 'alloweds', default_value=[])
    for allowed in alloweds:
      ip_protocol = self._GetJSONValue(allowed, 'IPProtocol')
      ports = self._GetJSONValue(allowed, 'ports', default_value='all')

      firewall_rules.append(f'ALLOW: {ip_protocol:s} {ports!s}')

    denieds = self._GetJSONValue(request, 'denieds', default_value=[])
    for denied in denieds:
      ip_protocol = self._GetJSONValue(denied, 'IPProtocol')
      ports = self._GetJSONValue(denied, 'ports', default_value='all')

      firewall_rules.append(f'DENY: {ip_protocol:s} {ports!s}')

    event_data.firewall_rules = firewall_rules or None

    # Service account specific attributes
    service_account = self._GetJSONValue(request, 'service_account')
    if service_account:
      event_data.service_account_display_name = self._GetJSONValue(
          service_account, 'display_name')

  def _ParseProtoPayloadServiceData(self, proto_payload, event_data):
    """Extracts information from the serviceData in the protoPayload value.

    Args:
      proto_payload (dict): JSON dictionary of the `protoPayload` value.
      event_data (GCPLogEventData): event data.
    """
    service_data = self._GetJSONValue(proto_payload, 'serviceData')
    if not service_data:
      return

    policy_delta = self._GetJSONValue(service_data, 'policyDelta')
    if not policy_delta:
      return

    policy_deltas = []

    binding_deltas = self._GetJSONValue(
        policy_delta, 'bindingDeltas', default_value=[])
    for binding_delta_value in binding_deltas:
      action = self._GetJSONValue(binding_delta_value, 'action') or 'N/A'
      member = self._GetJSONValue(binding_delta_value, 'member') or 'N/A'
      role = self._GetJSONValue(binding_delta_value, 'role') or 'N/A'

      policy_deltas.append(f'{action:s} {member:s} with role {role:s}')

    event_data.policy_deltas = policy_deltas or None

  def _ParseRecord(self, parser_mediator, json_dict):
    """Parses a Google Cloud (GCP) log record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary of the log record.
    """
    resource = self._GetJSONValue(json_dict, 'resource', default_value={})
    labels = self._GetJSONValue(resource, 'labels', default_value={})

    resource_labels = [f'{name:s}: {value!s}' for name, value in labels.items()]

    event_data = GCPLogEventData()
    event_data.log_name = self._GetJSONValue(json_dict, 'logName')
    event_data.recorded_time = self._ParseISO8601DateTimeString(
        parser_mediator, json_dict, 'timestamp')
    event_data.resource_labels = resource_labels or None
    event_data.severity = self._GetJSONValue(json_dict, 'severity')
    event_data.text_payload = self._GetJSONValue(json_dict, 'textPayload')

    self._ParseJSONPayload(json_dict, event_data)
    self._ParseProtoPayload(json_dict, event_data)

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, json_dict):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      json_dict (dict): JSON dictionary of the log record.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    log_name = json_dict.get('logName') or None
    timestamp = json_dict.get('timestamp') or None

    if None in (log_name, timestamp):
      return False

    date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()

    try:
      date_time.CopyFromStringISO8601(timestamp)
    except ValueError:
      return False

    return True


jsonl_parser.JSONLParser.RegisterPlugin(GCPLogJSONLPlugin)



"""An output module that writes event with geography data to a KML XML file.

The Keyhole Markup Language (KML) is an XML notation for expressing geographic
annotation and visualization within Internet-based, two-dimensional maps and
three-dimensional Earth browsers.
"""

import codecs

from xml.etree import ElementTree

from plaso.output import manager
from plaso.output import rawpy


class KMLOutputModule(rawpy.NativePythonOutputModule):
  """Output module for a Keyhole Markup Language (KML) XML file."""

  NAME = 'kml'
  DESCRIPTION = 'Saves events with geography data into a KML format.'

  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """
    latitude = field_values.get('latitude')
    longitude = field_values.get('longitude')
    if None in (latitude, longitude):
      return

    # TODO: make description_text KML values.
    description_text = self._GetString(field_values)

    placemark_xml_element = ElementTree.Element('Placemark')

    name_xml_element = ElementTree.SubElement(placemark_xml_element, 'name')
    name_xml_element.text = field_values['_event_identifier']

    description_xml_element = ElementTree.SubElement(
        placemark_xml_element, 'description')
    description_xml_element.text = f'{description_text:s}\n'

    point_xml_element = ElementTree.SubElement(placemark_xml_element, 'Point')

    coordinates_xml_element = ElementTree.SubElement(
        point_xml_element, 'coordinates')
    coordinates_xml_element.text = f'{longitude!s},{latitude!s}'

    # Note that ElementTree.tostring() will appropriately escape the input data.
    output_text = ElementTree.tostring(placemark_xml_element)

    output_text = codecs.decode(output_text, output_mediator.encoding)

    self.WriteText(output_text)

  def WriteFooter(self):
    """Writes the footer to the output."""
    self.WriteText('</Document></kml>')

  def WriteHeader(self, output_mediator):
    """Writes the header to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    self.WriteText((
        f'<?xml version="1.0" encoding="{output_mediator.encoding:s}"?>'
        f'<kml xmlns="http://www.opengis.net/kml/2.2"><Document>'))


manager.OutputManager.RegisterOutput(KMLOutputModule)



"""JSON-L parser plugin for Azure activity log files."""

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import jsonl_parser
from plaso.parsers.jsonl_plugins import interface


class AzureActivityLogEventData(events.EventData):
  """Azure activity log event data.

  Attributes:
    caller (str): Azure identity.
    client_ip (str): client IP address.
    correlation_identifier (str): Correlation identifier.
    event_data_identifier (str): Event data identifier.
    event_name (str): name of the event.
    level (str): log level.
    operation_identifier (str): Operation identifier.
    operation_name (str): operation name.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
    resource_group (str): resource group.
    resource_identifier (str): resource.
    resource_provider (str): API service.
    resource_type (str): resource type.
    subscription_identifier (str): subscription identifier.
    tenant_identifier (str): tenant identifier.
  """

  DATA_TYPE = 'azure:activitylog:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.caller = None
    self.client_ip = None
    self.correlation_identifier = None
    self.event_data_identifier = None
    self.event_name = None
    self.level = None
    self.operation_identifier = None
    self.operation_name = None
    self.recorded_time = None
    self.resource_group = None
    self.resource_identifier = None
    self.resource_provider = None
    self.resource_type = None
    self.subscription_identifier = None
    self.tenant_identifier = None


class AzureActivityLogJSONLPlugin(interface.JSONLPlugin):
  """JSON-L parser plugin for Azure activity log files."""

  NAME = 'azure_activity_log'
  DATA_FORMAT = 'Azure Activity Log'

  def _ParseRecord(self, parser_mediator, json_dict):
    """Parses an Azure activity log record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary of the log record.
    """
    date_time = None

    event_timestamp = self._GetJSONValue(json_dict, 'event_timestamp')
    if event_timestamp:
      try:
        date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()
        date_time.CopyFromStringISO8601(event_timestamp)
      except ValueError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'Unable to parse time string: {event_timestamp:s} with error: '
            f'{exception!s}')
        date_time = None

    event_name_json = self._GetJSONValue(
        json_dict, 'event_name', default_value={})
    http_request_json = self._GetJSONValue(
        json_dict, 'http_request', default_value={})
    resource_provider_name_json = self._GetJSONValue(
        json_dict, 'resource_provider_name', default_value={})
    resource_type_json = self._GetJSONValue(
        json_dict, 'resource_type', default_value={})
    operation_name_json = self._GetJSONValue(
        json_dict, 'operation_name', default_value={})

    event_data = AzureActivityLogEventData()
    event_data.caller = self._GetJSONValue(json_dict, 'caller')
    event_data.client_ip = self._GetJSONValue(
        http_request_json, 'client_ip_address')
    event_data.correlation_identifier = self._GetJSONValue(
        json_dict, 'correlation_id')
    event_data.event_data_identifier = self._GetJSONValue(
        json_dict, 'event_data_id')
    event_data.event_name = self._GetJSONValue(event_name_json, 'value')
    event_data.level = self._GetJSONValue(json_dict, 'level')
    event_data.operation_identifier = self._GetJSONValue(
        json_dict, 'operation_id')
    event_data.operation_name = self._GetJSONValue(operation_name_json, 'value')
    event_data.recorded_time = date_time
    event_data.resource_group = self._GetJSONValue(
        json_dict, 'resource_group_name')
    event_data.resource_identifier = self._GetJSONValue(
        json_dict, 'resource_id')
    event_data.resource_provider = self._GetJSONValue(
        resource_provider_name_json, 'value')
    event_data.resource_type = self._GetJSONValue(resource_type_json, 'value')
    event_data.subscription_identifier = self._GetJSONValue(
        json_dict, 'subscription_id')
    event_data.tenant_identifier = self._GetJSONValue(json_dict, 'tenant_id')

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, json_dict):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      json_dict (dict): JSON dictionary of the log record.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    event_timestamp = json_dict.get('event_timestamp') or None
    subscription_identifier = json_dict.get('subscription_id') or None

    if None in (event_timestamp, subscription_identifier):
      return False

    date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()

    try:
      date_time.CopyFromStringISO8601(event_timestamp)
    except ValueError:
      return False

    return True


jsonl_parser.JSONLParser.RegisterPlugin(AzureActivityLogJSONLPlugin)



"""Analysis result attribute containers."""

from acstore.containers import interface
from acstore.containers import manager


class BrowserSearchAnalysisResult(interface.AttributeContainer):
  """Browser search analysis plugin result container.

  Attributes:
    number_of_queries (int): number of times the search engine was queried.
    search_engine (str): search engine that was queried.
    search_term (str): term searched for.
  """

  CONTAINER_TYPE = 'browser_search_analysis_result'

  SCHEMA = {
      'number_of_queries': 'int',
      'search_engine': 'str',
      'search_term': 'str'}

  def __init__(
      self, number_of_queries=None, search_engine=None, search_term=None):
    """Initializes a browser search analysis plugin result container.

    Args:
      number_of_queries (Optional[int]): number of times the search engine was
          queried.
      search_engine (Optional[str]): search engine that was queried.
      search_term (Optional[str]): term searched for.
    """
    super().__init__()
    self.number_of_queries = number_of_queries
    self.search_engine = search_engine
    self.search_term = search_term


class ChromeExtensionAnalysisResult(interface.AttributeContainer):
  """Chrome extension analysis plugin result container.

  Attributes:
    extension (str): name of the Chrome extension.
    extension_identifier (str): identifier of the Chrome extension.
    username (str): name of a user that has installed the Chrome extension.
  """

  CONTAINER_TYPE = 'chrome_extension_analysis_result'

  SCHEMA = {
      'extension': 'str',
      'extension_identifier': 'str',
      'username': 'str'}

  def __init__(
      self, extension=None, extension_identifier=None, username=None):
    """Initializes a Chrome extension analysis plugin result container.

    Args:
      extension (Optional[str]): name of the Chrome extension.
      extension_identifier (Optional[str]): identifier of the Chrome extension.
      username (Optional[str]): name of a user that has installed the Chrome
          extension.
    """
    super().__init__()
    self.extension = extension
    self.extension_identifier = extension_identifier
    self.username = username


manager.AttributeContainersManager.RegisterAttributeContainers([
    BrowserSearchAnalysisResult, ChromeExtensionAnalysisResult])



"""File containing a Windows Registry plugin to parse the USB Device key.

Also see:
  https://learn.microsoft.com/en-us/windows-hardware/drivers/usbcon/
"""

from plaso.containers import events
from plaso.parsers import logger
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class WindowsUSBDeviceEventData(events.EventData):
  """Windows USB device event data attribute container.

  Attributes:
    key_path (str): Windows Registry key path.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    product (str): product of the USB device.
    serial (str): serial number of the USB device.
    subkey_name (str): name of the Windows Registry subkey.
    vendor (str): vendor of the USB device.
  """

  DATA_TYPE = 'windows:registry:usb'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.key_path = None
    self.last_written_time = None
    self.product = None
    self.serial = None
    # TODO: rename subkey_name to something that closer matches its purpose.
    self.subkey_name = None
    self.vendor = None


class USBPlugin(interface.WindowsRegistryPlugin):
  """USB Windows Registry plugin for last connection time."""

  NAME = 'windows_usb_devices'
  DATA_FORMAT = 'Windows USB device Registry data'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Enum\\USB')])

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    event_data = WindowsUSBDeviceEventData()
    event_data.key_path = registry_key.path

    for subkey in registry_key.GetSubkeys():
      event_data.subkey_name = subkey.name

      vendor_identification = None
      product_identification = None

      try:
        subkey_name_parts = subkey.name.split('&')
        if len(subkey_name_parts) >= 2:
          vendor_identification = subkey_name_parts[0]
          product_identification = subkey_name_parts[1]
      except ValueError as exception:
        logger.warning((
            f'Unable to split string: {subkey.name:s} with error: '
            f'{exception!s}'))

      event_data.vendor = vendor_identification
      event_data.product = product_identification

      for devicekey in subkey.GetSubkeys():
        # The last written time of the Registry key relates to the last time
        # an USB device was connected to the operating system.
        event_data.last_written_time = devicekey.last_written_time
        event_data.serial = devicekey.name

        parser_mediator.ProduceEventData(event_data)


winreg_parser.WinRegistryParser.RegisterPlugin(USBPlugin)



"""Heap to sort events in chronological order."""

import heapq


class EventHeap:
  """Event heap."""

  def __init__(self):
    """Initializes an event heap."""
    super().__init__()
    self._heap = []

  @property
  def number_of_events(self):
    """int: number of serialized events on the heap."""
    return len(self._heap)

  def PopEvent(self):
    """Pops an event from the heap.

    Returns:
      EventObject: event.
    """
    try:
      _, _, _, _, event = heapq.heappop(self._heap)
      return event

    except IndexError:
      return None

  def PopEvents(self):
    """Pops events from the heap.

    Yields:
      EventObject: event.
    """
    event = self.PopEvent()
    while event:
      yield event
      event = self.PopEvent()

  def PushEvent(self, event, event_index):
    """Pushes an event onto the heap.

    Args:
      event (EventObject): event.
      event_index (int): index of the event in the storage.
    """
    event_string = event.GetAttributeValuesString()
    heap_values = (
        event.timestamp, event.timestamp_desc, event_index, event_string, event)
    heapq.heappush(self._heap, heap_values)



"""SQLite parser plugin for MacOS LS quarantine events database files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


# TODO: describe more clearly what the data value contains.
class MacOSLSQuarantineEventData(events.EventData):
  """MacOS launch services quarantine event data.

  Attributes:
    agent (str): user agent that was used to download the file.
    data (bytes): data.
    downloaded_time (dfdatetime.DateTimeValues): date and time the file
        was downloaded.
    query (str): SQL query that was used to obtain the event data.
    url (str): original URL of the file.
  """

  DATA_TYPE = 'macos:lsquarantine:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.agent = None
    self.data = None
    self.downloaded_time = None
    self.query = None
    self.url = None


class MacOSLSQuarantinePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for MacOS LS quarantine events database files.

  The MacOS launch services (LS) quarantine database file is typically stored
  in: /Users/<username>/Library/Preferences/
      QuarantineEvents.com.apple.LaunchServices
  """

  NAME = 'ls_quarantine'
  DATA_FORMAT = (
      'MacOS launch services quarantine events database SQLite database file')

  REQUIRED_STRUCTURE = {
      'LSQuarantineEvent': frozenset([
          'LSQuarantineTimeStamp', 'LSQuarantineAgentName',
          'LSQuarantineOriginURLString', 'LSQuarantineDataURLString'])}

  QUERIES = [
      (('SELECT LSQuarantineTimeStamp AS Time, LSQuarantineAgentName AS Agent, '
        'LSQuarantineOriginURLString AS URL, LSQuarantineDataURLString AS Data '
        'FROM LSQuarantineEvent'), '_ParseLSQuarantineRow')]

  SCHEMAS = [{
      'LSQuarantineEvent': (
          'CREATE TABLE LSQuarantineEvent ( LSQuarantineEventIdentifier TEXT '
          'PRIMARY KEY NOT NULL, LSQuarantineTimeStamp REAL, '
          'LSQuarantineAgentBundleIdentifier TEXT, LSQuarantineAgentName '
          'TEXT, LSQuarantineDataURLString TEXT, LSQuarantineSenderName TEXT, '
          'LSQuarantineSenderAddress TEXT, LSQuarantineTypeNumber INTEGER, '
          'LSQuarantineOriginTitle TEXT, LSQuarantineOriginURLString TEXT, '
          'LSQuarantineOriginAlias BLOB )')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.CocoaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  def _ParseLSQuarantineRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses a launch services quarantine event row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = MacOSLSQuarantineEventData()
    event_data.agent = self._GetRowValue(query_hash, row, 'Agent')
    event_data.data = self._GetRowValue(query_hash, row, 'Data')
    event_data.downloaded_time = self._GetDateTimeRowValue(
        query_hash, row, 'Time')
    event_data.query = query
    event_data.url = self._GetRowValue(query_hash, row, 'URL')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(MacOSLSQuarantinePlugin)



"""An output module that saves events to OpenSearch for Timesketch."""

from plaso.output import logger
from plaso.output import manager
from plaso.output import shared_opensearch


class OpenSearchTimesketchOutputModule(
    shared_opensearch.SharedOpenSearchOutputModule):
  """Output module for Timesketch OpenSearch."""

  NAME = 'opensearch_ts'
  DESCRIPTION = (
      'Saves the events into an OpenSearch database for use '
      'with Timesketch.')

  MAPPINGS_FILENAME = 'plaso.mappings'
  MAPPINGS_PATH = '/etc/timesketch'

  def __init__(self):
    """Initializes an output module."""
    super().__init__()
    self._timeline_identifier = None

  def WriteFieldValues(self, output_mediator, field_values):
    """Writes field values to the output.

    Events are buffered in the form of documents and inserted to OpenSearch
    when the flush interval (threshold) has been reached.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.
    """
    event_document = {'index': {'_index': self._index_name}}

    # Add timeline_id on the event level. It is used in Timesketch to
    # support shared indices.
    field_values['__ts_timeline_id'] = self._timeline_identifier

    self._event_documents.append(event_document)
    self._event_documents.append(field_values)
    self._number_of_buffered_events += 1

    if self._number_of_buffered_events > self._flush_interval:
      self._FlushEvents()

  def GetMissingArguments(self):
    """Retrieves a list of arguments that are missing from the input.

    Returns:
      list[str]: names of arguments that are required by the module and have
          not been specified.
    """
    if not self._timeline_identifier:
      return ['timeline_id']
    return []

  def SetTimelineIdentifier(self, timeline_identifier):
    """Sets the timeline identifier.

    Args:
      timeline_identifier (int): timeline identifier.
    """
    self._timeline_identifier = timeline_identifier
    logger.info(f'Timeline identifier: {self._timeline_identifier:d}')

  def WriteHeader(self, output_mediator):
    """Connects to the OpenSearch server and creates the index.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    self._Connect()

    self._CreateIndexIfNotExists(self._index_name, self._mappings)


manager.OutputManager.RegisterOutput(
    OpenSearchTimesketchOutputModule,
    disabled=shared_opensearch.opensearchpy is None)



"""Redis-based attribute container store.

Only supports task storage at the moment.
"""

import ast
import json
import uuid

import redis  # pylint: disable=import-error

from acstore import interface
from acstore.containers import interface as containers_interface
from acstore.helpers import json_serializer as containers_json_serializer

from plaso.containers import events
from plaso.lib import definitions
from plaso.serializer import json_serializer
from plaso.storage import logger


class BaseRedisAttributeContainerStore(
    interface.AttributeContainerStoreWithReadCache):
  """Redis-based attribute container store.

  Attribute containers are stored as Redis hashes. All keys are prefixed with
  the session identifier to avoid collisions. Event identifiers are also stored
  in an index to enable sorting.

  Attributes:
    format_version (int): storage format version.
    serialization_format (str): serialization format.
  """

  _FORMAT_VERSION = 20230312

  DEFAULT_REDIS_URL = 'redis://127.0.0.1/0'

  def __init__(self):
    """Initializes a Redis attribute container store."""
    super().__init__()
    self._json_serializer = (
        containers_json_serializer.AttributeContainerJSONSerializer)
    self._redis_client = None
    self._session_identifier = None
    self._task_identifier = None

    self.format_version = self._FORMAT_VERSION
    self.serialization_format = definitions.SERIALIZER_FORMAT_JSON

  def _GetRedisHashName(self, container_type):
    """Retrieves the Redis hash name of the attribute container type.

    Args:
      container_type (str): container type.

    Returns:
      str: a Redis key name.
    """
    return (f'{self._session_identifier:s}-{self._task_identifier:s}-'
            f'{container_type:s}')

  def _RaiseIfNotReadable(self):
    """Raises if the attribute container store is not readable.

    Raises:
      OSError: when the attribute container store is closed.
    """
    if not self._redis_client:
      raise OSError('Unable to read, client not connected.')

  def _RaiseIfNotWritable(self):
    """Raises if the attribute container store is not writable.

    Raises:
      OSError: when the attribute container store is closed or read-only.
    """
    if not self._redis_client:
      raise OSError('Unable to write, client not connected.')

  @classmethod
  def _SetClientName(cls, redis_client, name):
    """Attempts to sets a Redis client name.

    This method ignores errors from the Redis server or exceptions
    indicating the method is missing, as setting the name is not a critical
    function, and it is not currently supported by the fakeredis test library.

    Args:
      redis_client (Redis): an open Redis client.
      name (str): name to set.
    """
    try:
      redis_client.client_setname(name)
    except redis.ResponseError as exception:
      logger.debug((
          f'Unable to set redis client name: {name:s} with error: '
          f'{exception!s}'))

  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """
    identifier = container.GetIdentifier()

    redis_hash_name = self._GetRedisHashName(container.CONTAINER_TYPE)
    redis_key = identifier.CopyToString()

    json_dict = self._json_serializer.ConvertAttributeContainerToJSON(container)

    try:
      json_string = json.dumps(json_dict)
    except TypeError as exception:
      raise OSError((
          f'Unable to serialize attribute container: '
          f'{container.CONTAINER_TYPE:s} with error: {exception!s}.'))

    if not json_string:
      raise OSError((
          f'Unable to serialize attribute container: '
          f'{container.CONTAINER_TYPE:s}'))

    self._redis_client.hset(redis_hash_name, key=redis_key, value=json_string)

  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """
    next_sequence_number = self._GetAttributeContainerNextSequenceNumber(
        container.CONTAINER_TYPE)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container.CONTAINER_TYPE, sequence_number=next_sequence_number)
    container.SetIdentifier(identifier)

    redis_hash_name = self._GetRedisHashName(container.CONTAINER_TYPE)
    redis_key = identifier.CopyToString()

    json_dict = self._json_serializer.ConvertAttributeContainerToJSON(container)

    try:
      json_string = json.dumps(json_dict)
    except TypeError as exception:
      raise OSError((
          f'Unable to serialize attribute container: '
          f'{container.CONTAINER_TYPE:s} with error: {exception!s}.'))

    if not json_string:
      raise OSError((
          f'Unable to serialize attribute container: '
          f'{container.CONTAINER_TYPE:s}'))

    self._redis_client.hsetnx(redis_hash_name, redis_key, json_string)

    self._CacheAttributeContainerByIndex(container, next_sequence_number - 1)

  def _WriteStorageMetadata(self):
    """Writes the storage metadata."""
    metadata = {
        'format_version': self._FORMAT_VERSION,
        'serialization_format': self.serialization_format}
    metadata_key = self._GetRedisHashName('metadata')

    for key, value in metadata.items():
      self._redis_client.hset(metadata_key, key=key, value=value)

  def Close(self):
    """Closes the store.

    Raises:
      OSError: if the store is already closed.
    """
    if not self._redis_client:
      raise OSError('Store already closed.')

    self._redis_client = None

  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    redis_key = identifier.CopyToString()

    json_string = self._redis_client.hget(redis_hash_name, redis_key)
    if not json_string:
      return None

    json_dict = json.loads(json_string)

    container = self._json_serializer.ConvertJSONToAttributeContainer(json_dict)
    container.SetIdentifier(identifier)
    return container

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    identifier = containers_interface.AttributeContainerIdentifier(
        name=container_type, sequence_number=index + 1)

    redis_hash_name = self._GetRedisHashName(container_type)
    redis_key = identifier.CopyToString()

    json_string = self._redis_client.hget(redis_hash_name, redis_key)
    if not json_string:
      return None

    json_dict = json.loads(json_string)

    container = self._json_serializer.ConvertJSONToAttributeContainer(json_dict)
    container.SetIdentifier(identifier)
    return container

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Yields:
      AttributeContainer: attribute container.
    """
    redis_hash_name = self._GetRedisHashName(container_type)

    if filter_expression:
      expression_ast = ast.parse(filter_expression, mode='eval')
      filter_expression = compile(expression_ast, '<string>', mode='eval')

    for redis_key, json_string in self._redis_client.hscan_iter(
        redis_hash_name):
      json_dict = json.loads(json_string)

      container = self._json_serializer.ConvertJSONToAttributeContainer(
          json_dict)
      # TODO: map filter expression to Redis native filter.
      if container.MatchesExpression(filter_expression):
        key = redis_key.decode('utf8')
        identifier = containers_interface.AttributeContainerIdentifier()
        identifier.CopyFromString(key)

        container.SetIdentifier(identifier)
        yield container

  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """
    redis_hash_name = self._GetRedisHashName(container_type)
    return self._redis_client.hlen(redis_hash_name)

  def HasAttributeContainers(self, container_type):
    """Determines if a store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """
    return self._attribute_container_sequence_numbers[container_type] > 0

  # pylint: disable=arguments-differ
  def Open(
      self, redis_client=None, session_identifier=None, task_identifier=None,
      url=None, **unused_kwargs):
    """Opens the store.

    Args:
      redis_client (Optional[Redis]): Redis client to query. If specified, no
          new client will be created. If no client is specified a new client
          will be opened connected to the Redis instance specified by 'url'.
      session_identifier (Optional[str]): identifier of the session.
      task_identifier (Optional[str]): unique identifier of the task the store
          will store containers for. If not specified, an identifier will be
          generated.
      url (Optional[str]): URL for a Redis database. If not specified, the
          DEFAULT_REDIS_URL will be used.

    Raises:
      OSError: if the store is already connected to a Redis instance.
    """
    if self._redis_client:
      raise OSError('Redis client already connected')

    if not redis_client:
      if not url:
        url = self.DEFAULT_REDIS_URL

      redis_client = redis.from_url(url=url, socket_timeout=60)

    self._redis_client = redis_client

    self._session_identifier = session_identifier or str(uuid.uuid4())
    self._task_identifier = task_identifier or str(uuid.uuid4())

    client_name = self._GetRedisHashName('')
    self._SetClientName(self._redis_client, client_name)

    metadata_key = self._GetRedisHashName('metadata')
    if not self._redis_client.exists(metadata_key):
      self._WriteStorageMetadata()


class RedisAttributeContainerStore(BaseRedisAttributeContainerStore):
  """Redis-based attribute container store.

  Attribute containers are stored as Redis hashes. All keys are prefixed with
  the session identifier to avoid collisions. Event identifiers are also stored
  in an index to enable sorting.
  """

  _CONTAINER_TYPE_EVENT = events.EventObject.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA = events.EventData.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA_STREAM = events.EventDataStream.CONTAINER_TYPE

  _EVENT_INDEX_NAME = 'sorted_event_identifier'

  def __init__(self):
    """Initializes a Redis attribute container store."""
    super().__init__()
    self._serializer = json_serializer.JSONAttributeContainerSerializer
    self._serializers_profiler = None

  def _DeserializeAttributeContainer(self, container_type, serialized_data):
    """Deserializes an attribute container.

    Args:
      container_type (str): attribute container type.
      serialized_data (bytes): serialized attribute container data.

    Returns:
      AttributeContainer: attribute container or None.

    Raises:
      OSError: if the serialized data cannot be decoded.
    """
    if not serialized_data:
      return None

    if self._serializers_profiler:
      self._serializers_profiler.StartTiming(container_type)

    try:
      serialized_string = serialized_data.decode('utf-8')
      attribute_container = self._serializer.ReadSerialized(serialized_string)

    except UnicodeDecodeError as exception:
      raise OSError(
          f'Unable to decode serialized data with error: {exception!s}')

    except (TypeError, ValueError) as exception:
      # TODO: consider re-reading attribute container with error correction.
      raise OSError(f'Unable to read serialized data with error: {exception!s}')

    finally:
      if self._serializers_profiler:
        self._serializers_profiler.StopTiming(container_type)

    return attribute_container

  def _SerializeAttributeContainer(self, attribute_container):
    """Serializes an attribute container.

    Args:
      attribute_container (AttributeContainer): attribute container.

    Returns:
      bytes: serialized attribute container.

    Raises:
      OSError: if the attribute container cannot be serialized.
    """
    if self._serializers_profiler:
      self._serializers_profiler.StartTiming(
          attribute_container.CONTAINER_TYPE)

    try:
      attribute_container_data = self._serializer.WriteSerialized(
          attribute_container)
      if not attribute_container_data:
        raise OSError((
            f'Unable to serialize attribute container: '
            f'{attribute_container.CONTAINER_TYPE:s}'))

    finally:
      if self._serializers_profiler:
        self._serializers_profiler.StopTiming(
            attribute_container.CONTAINER_TYPE)

    return attribute_container_data

  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """
    schema = self._GetAttributeContainerSchema(container.CONTAINER_TYPE)
    if schema:
      super()._WriteNewAttributeContainer(container)
    else:
      next_sequence_number = self._GetAttributeContainerNextSequenceNumber(
          container.CONTAINER_TYPE)

      identifier = containers_interface.AttributeContainerIdentifier(
          name=container.CONTAINER_TYPE, sequence_number=next_sequence_number)
      container.SetIdentifier(identifier)

      redis_hash_name = self._GetRedisHashName(container.CONTAINER_TYPE)
      redis_key = identifier.CopyToString()

      if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
        event_data_stream_identifier = container.GetEventDataStreamIdentifier()
        if event_data_stream_identifier:
          setattr(container, '_event_data_stream_identifier',
                  event_data_stream_identifier.sequence_number)

      serialized_data = self._SerializeAttributeContainer(container)
      self._redis_client.hsetnx(redis_hash_name, redis_key, serialized_data)

    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT:
      index_name = self._GetRedisHashName(self._EVENT_INDEX_NAME)

      identifier = container.GetIdentifier()
      redis_key = identifier.CopyToString()

      self._redis_client.zincrby(index_name, container.timestamp, redis_key)

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    schema = self._GetAttributeContainerSchema(container_type)
    if schema:
      return super().GetAttributeContainerByIndex(container_type, index)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container_type, sequence_number=index + 1)

    redis_hash_name = self._GetRedisHashName(container_type)
    redis_key = identifier.CopyToString()

    serialized_data = self._redis_client.hget(redis_hash_name, redis_key)
    if not serialized_data:
      return None

    container = self._DeserializeAttributeContainer(
        container_type, serialized_data)
    container.SetIdentifier(identifier)

    if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
      identifier = getattr(container, '_event_data_stream_identifier', None)
      if identifier:
        event_data_stream_identifier = (
            containers_interface.AttributeContainerIdentifier(
                name=self._CONTAINER_TYPE_EVENT_DATA_STREAM,
                sequence_number=identifier))
        container.SetEventDataStreamIdentifier(event_data_stream_identifier)

    return container

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Yields:
      AttributeContainer: attribute container.
    """
    schema = self._GetAttributeContainerSchema(container_type)
    if schema:
      yield from super().GetAttributeContainers(
          container_type, filter_expression=filter_expression)
    else:
      redis_hash_name = self._GetRedisHashName(container_type)
      for redis_key, serialized_data in self._redis_client.hscan_iter(
          redis_hash_name):
        container = self._DeserializeAttributeContainer(
            container_type, serialized_data)

        redis_key = redis_key.decode('utf-8')
        identifier = containers_interface.AttributeContainerIdentifier()
        identifier.CopyFromString(redis_key)

        if container.CONTAINER_TYPE == self._CONTAINER_TYPE_EVENT_DATA:
          identifier = getattr(container, '_event_data_stream_identifier', None)
          if identifier:
            event_data_stream_identifier = (
                containers_interface.AttributeContainerIdentifier(
                    name=self._CONTAINER_TYPE_EVENT_DATA_STREAM,
                    sequence_number=identifier))
            container.SetEventDataStreamIdentifier(event_data_stream_identifier)

        # TODO: map filter expression to Redis native filter.
        if container.MatchesExpression(filter_expression):
          yield container

  def GetSortedEvents(self, time_range=None):
    """Retrieves the events in increasing chronological order.

    Args:
      time_range (Optional[TimeRange]): This argument is not supported by the
          Redis store.

    Yields:
      EventObject: event.

    Raises:
      RuntimeError: if a time_range argument is specified.
    """
    event_index_name = self._GetRedisHashName(self._EVENT_INDEX_NAME)
    if time_range:
      raise RuntimeError('Not supported')

    for redis_key, _ in self._redis_client.zscan_iter(event_index_name):
      redis_key = redis_key.decode('utf-8')
      identifier = containers_interface.AttributeContainerIdentifier()
      identifier.CopyFromString(redis_key)

      yield self.GetAttributeContainerByIdentifier(
          self._CONTAINER_TYPE_EVENT, identifier)

  def SetSerializersProfiler(self, serializers_profiler):
    """Sets the serializers profiler.

    Args:
      serializers_profiler (SerializersProfiler): serializers profiler.
    """
    self._serializers_profiler = serializers_profiler



"""Imports for the JSON-L parser plugins."""

from plaso.parsers.jsonl_plugins import aws_cloudtrail_log
from plaso.parsers.jsonl_plugins import azure_activity_log
from plaso.parsers.jsonl_plugins import azure_application_gateway_log
from plaso.parsers.jsonl_plugins import docker_container_config
from plaso.parsers.jsonl_plugins import docker_container_log
from plaso.parsers.jsonl_plugins import docker_layer_config
from plaso.parsers.jsonl_plugins import gcp_log
from plaso.parsers.jsonl_plugins import ios_app_privacy
from plaso.parsers.jsonl_plugins import microsoft365_audit_log



"""SQLite-based storage writer."""

from plaso.lib import definitions
from plaso.storage import writer
from plaso.storage.sqlite import sqlite_file


class SQLiteStorageWriter(writer.StorageWriter):
  """SQLite-based storage writer."""

  def __init__(self, storage_type=definitions.STORAGE_TYPE_SESSION):
    """Initializes a storage writer.

    Args:
      storage_type (Optional[str]): storage type.
    """
    super().__init__()
    self._first_written_event_data_index = 0
    self._first_written_event_source_index = 0
    self._written_event_data_index = 0
    self._written_event_source_index = 0

  def GetFirstWrittenEventData(self):
    """Retrieves the first event data that was written after open.

    Using GetFirstWrittenEventData and GetNextWrittenEventData newly
    added event data can be retrieved in order of addition.

    Returns:
      EventData: event data or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_data = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_DATA, self._first_written_event_data_index)

    if event_data:
      self._written_event_data_index = self._first_written_event_data_index + 1
    return event_data

  def GetNextWrittenEventData(self):
    """Retrieves the next event data that was written after open.

    Returns:
      EventData: event data or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_data = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_DATA, self._written_event_data_index)
    if event_data:
      self._written_event_data_index += 1
    return event_data

  def GetFirstWrittenEventSource(self):
    """Retrieves the first event source that was written after open.

    Using GetFirstWrittenEventSource and GetNextWrittenEventSource newly
    added event sources can be retrieved in order of addition.

    Returns:
      EventSource: event source or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_source = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_SOURCE,
        self._first_written_event_source_index)

    if event_source:
      self._written_event_source_index = (
          self._first_written_event_source_index + 1)
    return event_source

  def GetNextWrittenEventSource(self):
    """Retrieves the next event source that was written after open.

    Returns:
      EventSource: event source or None if there are no newly written ones.

    Raises:
      OSError: when the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    event_source = self._store.GetAttributeContainerByIndex(
        self._CONTAINER_TYPE_EVENT_SOURCE, self._written_event_source_index)
    if event_source:
      self._written_event_source_index += 1
    return event_source

  # pylint: disable=arguments-differ
  def Open(self, path=None, **unused_kwargs):
    """Opens the storage writer.

    Args:
      path (Optional[str]): path to the output SQLite database.

    Raises:
      OSError: if the storage writer is already opened.
    """
    if self._store:
      raise OSError('Storage writer already opened.')

    self._store = sqlite_file.SQLiteStorageFile()

    if self._serializers_profiler:
      self._store.SetSerializersProfiler(self._serializers_profiler)

    if self._storage_profiler:
      self._store.SetStorageProfiler(self._storage_profiler)

    self._store.Open(path=path, read_only=False)

    number_of_containers = self._store.GetNumberOfAttributeContainers(
        self._CONTAINER_TYPE_EVENT_DATA)
    self._first_written_event_data_index = number_of_containers
    self._written_event_data_index = self._first_written_event_data_index

    number_of_containers = self._store.GetNumberOfAttributeContainers(
        self._CONTAINER_TYPE_EVENT_SOURCE)
    self._first_written_event_source_index = number_of_containers
    self._written_event_source_index = self._first_written_event_source_index



"""Plist parser plugin for Apple iOS Mobile Backup plist files.

The plist contains history of opened applications in the Mobile Backup.
"""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class IOSMobileBackupEventData(events.EventData):
  """Apple iOS Mobile Backup event data.

  Attributes:
    account_enabled_time (dfdatetime.DateTimeValues): date and time the mobile
        backup account was enabled.
    backup_state (int): backup state.
    backup_state_time (dfdatetime.DateTimeValues): date and time if the backup
        state.
  """

  DATA_TYPE = 'ios:mobile_backup:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.account_enabled_time = None
    self.backup_state = None
    self.backup_state_time = None


class IOSMobileBackupStateEventData(events.EventData):
  """Apple iOS Mobile Backup state event data.

  Attributes:
    backup_time (dfdatetime.DateTimeValues): date and time of the backup.
    description (str): localized description of the backup state.
  """

  DATA_TYPE = 'ios:mobile_backup:state'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.backup_time = None
    self.description = None


class IOSMobileBackupPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Apple iOS Mobile Backup plist files."""

  NAME = 'ios_mobile_backup'
  DATA_FORMAT = 'Apple iOS Mobile Backup plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.MobileBackup.plist')])

  PLIST_KEYS = frozenset(['AccountEnabledDate', 'BackupStateInfo'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, top_level=None, **unused_kwargs):
    """Extract Mobile Backup history entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    event_data = IOSMobileBackupEventData()
    event_data.account_enabled_time = self._GetDateTimeValueFromPlistKey(
        top_level, 'AccountEnabledDate')

    # TODO: add support for LastOnConditionEvents
    # TODO: add support for RemoteConfigurationExpiration

    backup_state_info = top_level.get('BackupStateInfo')
    if backup_state_info:
      event_data.backup_state = backup_state_info.get('state')
      # TODO: determine what this time represents.
      event_data.backup_state_time = self._GetDateTimeValueFromPlistKey(
          backup_state_info, 'date')

    parser_mediator.ProduceEventData(event_data)

    if backup_state_info:
      for error in backup_state_info.get('errors', []):
        event_data = IOSMobileBackupStateEventData()
        event_data.description = error.get('localizedDescription')
        # TODO: determine if this is the start or end time of the backup.
        event_data.backup_time = self._GetDateTimeValueFromPlistKey(
            error, 'date')

        parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(IOSMobileBackupPlistPlugin)



"""Redis-based storage writer."""

from plaso.storage import writer
from plaso.storage.redis import redis_store


class RedisStorageWriter(writer.StorageWriter):
  """Redis-based storage writer."""

  # pylint: disable=redundant-returns-doc,useless-return
  def GetFirstWrittenEventData(self):
    """Retrieves the first event data that was written after open.

    Using GetFirstWrittenEventData and GetNextWrittenEventData newly
    added event data can be retrieved in order of addition.

    Returns:
      EventData: None as there are no newly written event data.

    Raises:
      OSError: if the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    return None

  # pylint: disable=redundant-returns-doc,useless-return
  def GetNextWrittenEventData(self):
    """Retrieves the next event data that was written after open.

    Returns:
      EventData: None as there are no newly written event data.

    Raises:
      OSError: if the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    return None

  # pylint: disable=redundant-returns-doc,useless-return
  def GetFirstWrittenEventSource(self):
    """Retrieves the first event source that was written after open.

    Using GetFirstWrittenEventSource and GetNextWrittenEventSource newly
    added event sources can be retrieved in order of addition.

    Returns:
      EventSource: None as there are no newly written event sources.

    Raises:
      OSError: if the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    return None

  # pylint: disable=redundant-returns-doc,useless-return
  def GetNextWrittenEventSource(self):
    """Retrieves the next event source that was written after open.

    Returns:
      EventSource: None as there are no newly written event sources.

    Raises:
      OSError: if the storage writer is closed.
    """
    if not self._store:
      raise OSError('Unable to read from closed storage writer.')

    return None

  # pylint: disable=arguments-differ
  def Open(
      self, redis_client=None, session_identifier=None, task_identifier=None,
      **unused_kwargs):
    """Opens the storage writer.

    Args:
      redis_client (Optional[Redis]): Redis client to query. If specified, no
          new client will be created. If no client is specified a new client
          will be opened connected to the Redis instance specified by 'url'.
      session_identifier (Optional[str]): session identifier.
      task_identifier (Optional[str]): task identifier.

    Raises:
      OSError: if the storage writer is already opened.
    """
    if self._store:
      raise OSError('Storage writer already opened.')

    self._store = redis_store.RedisAttributeContainerStore()

    if self._serializers_profiler:
      self._store.SetSerializersProfiler(self._serializers_profiler)

    if self._storage_profiler:
      self._store.SetStorageProfiler(self._storage_profiler)

    self._store.Open(
        redis_client=redis_client, session_identifier=session_identifier,
        task_identifier=task_identifier)



"""Parser for Windows XML EventLog (EVTX) files."""

import pyevtx

from dfdatetime import filetime as dfdatetime_filetime

from plaso.containers import events
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class WinEvtxRecordEventData(events.EventData):
  """Windows XML EventLog (EVTX) record event data.

  Attributes:
    creation_time (dfdatetime.DateTimeValues): event record creation date
        and time.
    computer_name (str): computer name stored in the event record.
    event_identifier (int): event identifier.
    event_level (int): event level.
    event_version (int): event version.
    message_identifier (int): event message identifier.
    offset (int): offset of the EVTX record relative to the start of the file,
        from which the event data was extracted.
    provider_identifier (str): identifier of the EventLog provider.
    record_number (int): event record number.
    recovered (bool): True if the record was recovered.
    source_name (str): name of the event source.
    strings (list[str]): event strings.
    user_sid (str): user security identifier (SID) stored in the event record.
    written_time (dfdatetime.DateTimeValues): event record written date and
        time.
    xml_string (str): XML representation of the event.
  """

  DATA_TYPE = 'windows:evtx:record'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.creation_time = None
    self.computer_name = None
    self.event_identifier = None
    self.event_level = None
    self.event_version = None
    self.message_identifier = None
    self.offset = None
    self.provider_identifier = None
    self.record_number = None
    self.recovered = None
    self.source_name = None
    self.strings = None
    self.user_sid = None
    self.written_time = None
    self.xml_string = None


class WinEvtxParser(interface.FileObjectParser):
  """Parses Windows XML EventLog (EVTX) files."""

  _INITIAL_FILE_OFFSET = None

  NAME = 'winevtx'
  DATA_FORMAT = 'Windows XML EventLog (EVTX) file'

  def _GetEventDataFromRecord(
      self, parser_mediator, record_index, evtx_record, recovered=False):
    """Extract data from a Windows XML EventLog (EVTX) record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      record_index (int): event record index.
      evtx_record (pyevtx.record): event record.
      recovered (Optional[bool]): True if the record was recovered.

    Return:
      WinEvtxRecordEventData: event data.
    """
    event_data = WinEvtxRecordEventData()

    try:
      event_data.record_number = evtx_record.identifier
    except OverflowError as exception:
      warning_message = (
          f'unable to read record identifier from event record: '
          f'{record_index:d} with error: {exception!s}')
      if recovered:
        parser_mediator.ProduceRecoveryWarning(warning_message)
      else:
        parser_mediator.ProduceExtractionWarning(warning_message)

    try:
      event_identifier = evtx_record.event_identifier
    except OverflowError as exception:
      warning_message = (
          f'unable to read event identifier from event record: '
          f'{record_index:d} with error: {exception!s}')
      if recovered:
        parser_mediator.ProduceRecoveryWarning(warning_message)
      else:
        parser_mediator.ProduceExtractionWarning(warning_message)

      event_identifier = None

    try:
      event_identifier_qualifiers = evtx_record.event_identifier_qualifiers
    except OverflowError as exception:
      warning_message = (
          f'unable to read event identifier qualifiers from event record: '
          f'{record_index:d} with error: {exception!s}')
      if recovered:
        parser_mediator.ProduceRecoveryWarning(warning_message)
      else:
        parser_mediator.ProduceExtractionWarning(warning_message)

      event_identifier_qualifiers = None

    event_data.offset = evtx_record.offset
    event_data.recovered = recovered

    if event_identifier is not None:
      event_data.event_identifier = event_identifier

      event_data.message_identifier = event_identifier
      if event_identifier_qualifiers is not None:
        event_data.message_identifier |= event_identifier_qualifiers << 16

    if evtx_record.provider_identifier:
      event_data.provider_identifier = evtx_record.provider_identifier.lower()

    event_data.event_level = evtx_record.event_level
    event_data.event_version = evtx_record.event_version
    event_data.source_name = evtx_record.source_name

    # Computer name is the value stored in the event record and does not
    # necessarily correspond with the actual hostname.
    event_data.computer_name = evtx_record.computer_name
    event_data.user_sid = evtx_record.user_security_identifier

    event_data.strings = list(evtx_record.strings)

    event_data.xml_string = evtx_record.xml_string

    return event_data

  def _ParseRecord(
      self, parser_mediator, record_index, evtx_record, recovered=False):
    """Extract data from a Windows XML EventLog (EVTX) record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      record_index (int): event record index.
      evtx_record (pyevtx.record): event record.
      recovered (Optional[bool]): True if the record was recovered.
    """
    event_data = self._GetEventDataFromRecord(
        parser_mediator, record_index, evtx_record, recovered=recovered)

    try:
      creation_time = evtx_record.get_creation_time_as_integer()
    except OverflowError as exception:
      warning_message = (
          f'unable to read creation time from event record: {record_index:d} '
          f'with error: {exception!s}')
      if recovered:
        parser_mediator.ProduceRecoveryWarning(warning_message)
      else:
        parser_mediator.ProduceExtractionWarning(warning_message)

      creation_time = None

    if creation_time:
      event_data.creation_time = dfdatetime_filetime.Filetime(
          timestamp=creation_time)

    try:
      written_time = evtx_record.get_written_time_as_integer()
    except OverflowError as exception:
      warning_message = (
          f'unable to read written time from event record: {record_index:d} '
          f'with error: {exception!s}')
      if recovered:
        parser_mediator.ProduceRecoveryWarning(warning_message)
      else:
        parser_mediator.ProduceExtractionWarning(warning_message)

      written_time = None

    if written_time:
      event_data.written_time = dfdatetime_filetime.Filetime(
          timestamp=written_time)

    parser_mediator.ProduceEventData(event_data)

  def _ParseRecords(self, parser_mediator, evtx_file):
    """Parses Windows XML EventLog (EVTX) records.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      evtx_file (pyevtx.file): Windows XML EventLog (EVTX) file.
    """
    # To handle errors when parsing a Windows XML EventLog (EVTX) file in the
    # most granular way the following code iterates over every event record.
    # The call to evt_file.get_record() and access to members of evt_record
    # should be called within a try-except.

    for record_index in range(evtx_file.number_of_records):
      if parser_mediator.abort:
        break

      try:
        evtx_record = evtx_file.get_record(record_index)
        self._ParseRecord(parser_mediator, record_index, evtx_record)

      except OSError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to parse event record: {record_index:d} with error: '
            f'{exception!s}')

    for record_index in range(evtx_file.number_of_recovered_records):
      if parser_mediator.abort:
        break

      try:
        evtx_record = evtx_file.get_recovered_record(record_index)
        self._ParseRecord(
            parser_mediator, record_index, evtx_record, recovered=True)

      except OSError as exception:
        parser_mediator.ProduceRecoveryWarning(
            f'unable to parse recovered event record: {record_index:d} with '
            f'error: {exception!s}')

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(b'ElfFile\x00', offset=0)
    return format_specification

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a Windows XML EventLog (EVTX) file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): a file-like object.
    """
    code_page = parser_mediator.GetCodePage()

    evtx_file = pyevtx.file()
    evtx_file.set_ascii_codepage(code_page)

    try:
      evtx_file.open_file_object(file_object)
    except OSError as exception:
      parser_mediator.ProduceExtractionWarning(
          f'unable to open file with error: {exception!s}')
      return

    try:
      self._ParseRecords(parser_mediator, evtx_file)
    finally:
      evtx_file.close()


manager.ParsersManager.RegisterParser(WinEvtxParser)



"""Plug-in to collect information about the Motherboard and BIOS."""

from plaso.containers import events
from plaso.parsers import winreg_parser
from plaso.parsers.winreg_plugins import interface


class WindowsRegistryMotherboardInfoEventData(events.EventData):
  """Windows Motherboard Info event data attribute container.

  Attributes:
    bios_release_date (str): Date of release of the installed BIOS.
    bios_version (str): Version of installed BIOS.
    key_path (str): Windows Registry key path.
    last_written_time (dfdatetime.DateTimeValues): entry last written date
        and time.
    motherboard_manufacturer (str): Motherboard manafacturer name.
    motherboard_model (str): Name of the specific motherboard model.
  """

  DATA_TYPE = 'windows:registry:motherboard_info'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.bios_release_date = None
    self.bios_version = None
    self.key_path = None
    self.last_written_time = None
    self.motherboard_manufacturer = None
    self.motherboard_model = None


class MotherboardInfoPlugin(interface.WindowsRegistryPlugin):
  """Plug-in to collect information about the Motherboard and BIOS."""

  NAME = 'motherboard_info'
  DATA_FORMAT = 'Motherboard Info Registry data'

  FILTERS = frozenset([
    interface.WindowsRegistryKeyPathFilter(
      'HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\'
      'SystemInformation')])

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    event_data = WindowsRegistryMotherboardInfoEventData()
    event_data.motherboard_manufacturer = self._GetValueFromKey(
      registry_key, 'SystemManufacturer')
    event_data.motherboard_model = self._GetValueFromKey(
      registry_key, 'SystemProductName')
    event_data.bios_release_date = self._GetValueFromKey(
      registry_key, 'BIOSReleaseDate')
    event_data.bios_version = self._GetValueFromKey(
      registry_key, 'BIOSVersion')
    event_data.key_path = registry_key.path
    event_data.last_written_time = registry_key.last_written_time

    parser_mediator.ProduceEventData(event_data)


winreg_parser.WinRegistryParser.RegisterPlugin(MotherboardInfoPlugin)



"""SQLite parser plugin for MacOS application usage database files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class MacOSApplicationUsageEventData(events.EventData):
  """MacOS application usage event data.

  Attributes:
    application (str): name of the application.
    application_version (str): version of the application.
    bundle_identifier (str): bundle identifier of the application.
    count (int): number of occurances of the event.
    event (str): event.
    last_used_time (dfdatetime.DateTimeValues): last date and time
        the application was last used.
    query (str): SQL query that was used to obtain the event data.
  """

  DATA_TYPE = 'macos:application_usage:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.application = None
    self.application_version = None
    self.bundle_identifier = None
    self.count = None
    self.event = None
    self.last_used_time = None
    self.query = None


class MacOSApplicationUsagePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for MacOS application usage database files.

  The MacOS application usage database is typically stored in:
  /var/db/application_usage.sqlite

  Application usage is a SQLite database that logs down entries triggered by
  NSWorkspaceWillLaunchApplicationNotification and
  NSWorkspaceDidTerminateApplicationNotification NSWorkspace notifications by
  crankd.

  More information can be found here:
  https://github.com/google/macops/blob/master/crankd/ApplicationUsage.py
  """

  NAME = 'appusage'
  DATA_FORMAT = (
      'MacOS application usage SQLite database (application_usage.sqlite) file')

  REQUIRED_STRUCTURE = {
      'application_usage': frozenset([
          'last_time', 'event', 'bundle_id', 'app_version', 'app_path',
          'number_times'])}

  QUERIES = [(
      ('SELECT last_time, event, bundle_id, app_version, app_path, '
       'number_times FROM application_usage'), '_ParseApplicationUsageRow')]

  SCHEMAS = [{
      'application_usage': (
          'CREATE TABLE application_usage (event TEXT, bundle_id TEXT, '
          'app_version TEXT, app_path TEXT, last_time INTEGER DEFAULT 0, '
          'number_times INTEGER DEFAULT 0, PRIMARY KEY (event, bundle_id))')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTime(timestamp=timestamp)

  def _ParseApplicationUsageRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses an application usage row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = MacOSApplicationUsageEventData()
    event_data.application = self._GetRowValue(query_hash, row, 'app_path')
    event_data.application_version = self._GetRowValue(
        query_hash, row, 'app_version')
    event_data.bundle_identifier = self._GetRowValue(
        query_hash, row, 'bundle_id')
    event_data.count = self._GetRowValue(query_hash, row, 'number_times')
    event_data.event = self._GetRowValue(query_hash, row, 'event')
    event_data.last_used_time = self._GetDateTimeRowValue(
        query_hash, row, 'last_time')
    event_data.query = query

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(MacOSApplicationUsagePlugin)



"""This file imports Python modules that register analysis plugins."""

try:
  from plaso.analysis import bloom
except ImportError:
  pass

from plaso.analysis import browser_search
from plaso.analysis import chrome_extension
from plaso.analysis import nsrlsvr
from plaso.analysis import sessionize
from plaso.analysis import tagging
from plaso.analysis import test_memory
from plaso.analysis import unique_domains_visited
from plaso.analysis import viper
from plaso.analysis import virustotal



"""Text parser plugin for Android logcat files.

Android logcat can have a number of output formats. This parser supports:
* 'threadtime' format
* 'time' format

The log file format is:
date time PID-TID/package priority/tag: message

For example:
12-10 13:02:50.071 1901-4229/com.google.android.gms V/AuthZen: Handling
delegate intent.

Where priority is:
V: Verbose (lowest priority)
D: Debug
I: Info
W: Warning
E: Error
A: Assert

In addition, support for the format modifiers:
* date with a year
* user identifier (uid)
* microseconds fraction of second precision (usec)
* time zone offset

Also see:
  https://developer.android.com/studio/debug/logcat
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import dateless_helper
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class AndroidLogcatEventData(events.EventData):
  """Android logcat event data.

  Attributes:
    component_tag (str): the tag that indicates the system component from which
        the logcat line originates.
    file_offset (int): the file offset of where the log message was parsed.
    message (str): the log message.
    pid (int): process identifier (PID) that created the logcat line.
    priority (str): a character in the set {V, D, I, W, E, F, S}, which is
        ordered from lowest to highest priority.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
    thread_identifier (int): thread identifier (TID) that created the logcat
        line.
    user_identifier (int): the user identifier (UID) or Android ID of
        the logged process.
  """

  DATA_TYPE = 'android:logcat'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.component_tag = None
    self.file_offset = None
    self.message = None
    self.pid = None
    self.priority = None
    self.recorded_time = None
    self.thread_identifier = None
    self.user_identifier = None


class AndroidLogcatTextPlugin(
    interface.TextPlugin, dateless_helper.DateLessLogFormatHelper):
  """Text parser plugin for Android logcat files."""

  NAME = 'android_logcat'
  DATA_FORMAT = 'Android logcat file'

  ENCODING = 'utf-8'

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _MONTH_DAY = (
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS)

  _YEAR_MONTH_DAY = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') + _TWO_DIGITS)

  _FRACTION_OF_SECOND = (
      pyparsing.Word(pyparsing.nums, exact=3) ^
      pyparsing.Word(pyparsing.nums, exact=6))

  # Date and time values are formatted as:
  # 01-02 01:02:04.156 (year-less)
  # 2022-01-02 01:20:03.171
  # 2022-01-02 11:44:23.183801
  _DATE_TIME = (
      pyparsing.Or([_YEAR_MONTH_DAY, _MONTH_DAY]) +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress('.') +
      _FRACTION_OF_SECOND)

  _TIME_ZONE_OFFSET = (
      pyparsing.Word('+-', exact=1) + _TWO_DIGITS + _TWO_DIGITS)

  _PID_AND_THREAD_IDENTIFIER = (
      _INTEGER.set_results_name('pid') +
      _INTEGER.set_results_name('thread_identifier'))

  _USER_PID_AND_THREAD_IDENTIFIER = (
      _INTEGER.set_results_name('user_identifier') +
      _PID_AND_THREAD_IDENTIFIER)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _BEGINNING_LINE = (
      pyparsing.Suppress('--------- beginning of ') +
      pyparsing.one_of(['events', 'kernel', 'main', 'radio', 'system']) +
      _END_OF_LINE)

  _THREADTIME_LINE_BODY = (
      pyparsing.Or([
          _USER_PID_AND_THREAD_IDENTIFIER, _PID_AND_THREAD_IDENTIFIER]) +
      pyparsing.Word('VDIWEFS', exact=1).set_results_name('priority') +
      pyparsing.Optional(pyparsing.Word(
          pyparsing.printables + ' ', excludeChars=':').set_results_name(
              'tag')))

  _TIME_LINE_BODY = (
      pyparsing.Word('VDIWEFS', exact=1).set_results_name('priority') +
      pyparsing.Suppress('/') +
      pyparsing.Word(
          pyparsing.printables + ' ', excludeChars='(').set_results_name(
              'tag') +
      pyparsing.Suppress('(') +
      pyparsing.Or([
          _INTEGER.set_results_name('pid'),
          (_INTEGER.set_results_name('user_identifier') +
           pyparsing.Suppress(':') + _INTEGER.set_results_name('pid'))]) +
      pyparsing.Suppress(')'))

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.Optional(_TIME_ZONE_OFFSET).set_results_name(
          'time_zone_offset') +
      (_THREADTIME_LINE_BODY ^ _TIME_LINE_BODY) +
      pyparsing.Suppress(': ') +
      pyparsing.restOfLine().set_results_name('message') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('beginning_line', _BEGINNING_LINE),
      ('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _BEGINNING_LINE ^ _LOG_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key != 'beginning_line':
      event_data = AndroidLogcatEventData()
      event_data.component_tag = self._GetStringValueFromStructure(
          structure, 'tag')
      event_data.file_offset = self._current_offset
      event_data.message = self._GetValueFromStructure(structure, 'message')
      event_data.pid = self._GetValueFromStructure(structure, 'pid')
      event_data.priority = self._GetValueFromStructure(structure, 'priority')
      event_data.recorded_time = self._ParseTimeElements(structure)
      event_data.thread_identifier = self._GetValueFromStructure(
          structure, 'thread_identifier')
      event_data.user_identifier = self._GetValueFromStructure(
          structure, 'user_identifier')

      parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, structure):
    """Parses date and time elements of a log line.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      time_elements_structure = self._GetValueFromStructure(
          structure, 'date_time')

      has_year = len(time_elements_structure) == 7

      if has_year:
        (year, month, day_of_month, hours, minutes, seconds,
         fraction_of_second_string) = time_elements_structure
      else:
        (month, day_of_month, hours, minutes, seconds,
         fraction_of_second_string) = time_elements_structure

        self._UpdateYear(month)

        year = self._GetRelativeYear()

      time_zone_offset = self._GetValueFromStructure(
          structure, 'time_zone_offset')
      if time_zone_offset:
        time_zone_sign, time_zone_hours, time_zone_minutes = time_zone_offset

        time_zone_offset = (time_zone_hours * 60) + time_zone_minutes
        if time_zone_sign == '-':
          time_zone_offset *= -1

      fraction_of_second = int(fraction_of_second_string, 10)

      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds,
          fraction_of_second)

      if len(fraction_of_second_string) == 3:
        date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
            is_delta=(not has_year), time_elements_tuple=time_elements_tuple,
            time_zone_offset=time_zone_offset)
      else:
        date_time = dfdatetime_time_elements.TimeElementsInMicroseconds(
            is_delta=(not has_year), time_elements_tuple=time_elements_tuple,
            time_zone_offset=time_zone_offset)

      if time_zone_offset is None:
        date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    self._SetEstimatedYear(parser_mediator)

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')
    if time_elements_structure:
      try:
        self._ParseTimeElements(time_elements_structure)
      except errors.ParseError:
        return False

    return True


text_parser.TextLogParser.RegisterPlugin(AndroidLogcatTextPlugin)
